"""Module containing occurrence data wranglers for modifying point data."""
from lmpy.data_wrangling.common.accepted_name_wrangler import (
    _AcceptedNameWrangler,
    resolve_names_gbif,
)
from lmpy.data_wrangling.tree.base import _TreeDataWrangler


# .....................................................................................
class AcceptedNameTreeWrangler(_TreeDataWrangler, _AcceptedNameWrangler):
    """Modifies tree columns to update taxon names to the "accepted" names."""
    name = 'AcceptedNameTreeWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        name_map=None,
        name_resolver=None,
        purge_failures=True,
        **params
    ):
        """Constructor for AcceptedNameTreeModifier class.

        Args:
            name_map (dict): A map of original name to accepted name.
            name_resolver (str or Method): If provided, use this method for getting new
                accepted names.  If set to 'gbif', use GBIF name resolution.
            purge_failures (bool): Should failures be purged from the tree.
            **params (dict): Keyword parameters to pass to _TreeDataWrangler.
        """
        if isinstance(name_resolver, str) and name_resolver.lower() == 'gbif':
            name_resolver = resolve_names_gbif
        _AcceptedNameWrangler.__init__(
            self,
            name_map=name_map,
            name_resolver=name_resolver
        )
        _TreeDataWrangler.__init__(self, **params)

        self.purge_failures = purge_failures

    # .......................
    def wrangle_tree(self, tree):
        """Wrangle a tree.

        Args:
            tree (TreeWrapper): A tree to wrangle.

        Returns:
            TreeWrapper: The tree with accepted taxon names.
        """
        failures = []
        for taxon in tree.taxon_namespace:
            acc_name = self.resolve_names([taxon.label])[taxon.label]
            if acc_name is None:
                failures.append(taxon)
                taxon.label = ''
            elif taxon.label != acc_name:
                taxon.label = acc_name
                # report
                self._report_tip(modified=True)

        # Should we purge?
        if self.purge_failures and len(failures) > 0:
            for _ in failures:
                self._report_tip(purged=True)

            tree.prune_taxa(failures)
            tree.purge_taxon_namespace()

        return tree



"""Module for performing a MetaCommunity Phylogenetics Analysis.

See:
    Leibold, M.A., E.P. Economo and P.R. Peres-Neto. 2010. Metacommunity
         phylogenetics: separating the roles of environmental filters and
         historical biogeography. Ecology letters 13: 1290-1299.
"""
from concurrent.futures import ThreadPoolExecutor as ExecutorClass
from functools import partial
import threading

import numpy as np

from lmpy import Matrix

# Note: This lock is used when calculating beta to keep memory usage down
lock = threading.Lock()

# Note: This will depend on the executor class used and is subject to tuning.
#    Threads should have a higher level of concurrency than processes.
CONCURRENCY_FACTOR = 5


# .............................................................................
def _beta_helper(mtx1, mtx2, weights):
    """This helper function avoids creating large temporary matrices.

    Args:
        mtx1 (Matrix): A (n [sites] by i [predictors]) standardized matrix.
        mtx2 (Matrix): A (n [sites] by i [predictors]) standardized matrix.
        weights (Matrix): A (n [sites]) array of site weights.

    Returns:
        Matrix: The beta matrix created by the helper.
    """
    _, num_predictors = mtx1.shape
    _, num_k = mtx2.shape
    out_mtx = np.empty((num_predictors, num_k))
    for i in range(num_predictors):
        for j in range(i, num_k):
            out_mtx[i, j] = np.sum(mtx1[:, i] * weights * mtx2[:, j])
    return out_mtx


# .............................................................................
def _calculate_beta(pred_std, weights, phylo_std, use_lock=True):
    """Calculates the regression model (beta) for the provided inputs.

    Args:
        pred_std (Matrix): A standardized predictor matrix
            (n [sites] by i [predictors]).
        weights (Matrix): A matrix of site weights (n by n).
        phylo_std (Matrix): A standardized phylo matrix (n by k [nodes]).
        use_lock (boolean): If true, use a lock when performing the computation
            to save on overall memory usage.  This is useful for large, or
            full, predictor matrices but can be skipped for single columns to
            improve performance.

    Note:
        * The computation is::
            beta = (M_T.W.M)^-1.M_T.W.P
        * M is the predictor matrix
        * M_T is the transverse of the predictor matrix
        * W is the weights column
        * P is the phylo matrix
        * "^-1" is the inverse of the matrix
        * Locking is available to prevent many threads / subprocesses from
            performing memory intensive computations concurrently and
            overwhelming the system.

    Todo:
        * Update documentation so that note shows symbols in equation

    Returns:
        Matrix: An (i by k) numpy ndarray, where i is the number of predictors in
            pred_std and k is the number of nodes in phylo_std.
    """
    if use_lock:  # pragma: no cover
        lock.acquire()
    temp1 = _beta_helper(pred_std, pred_std, weights)
    tmp1_inv = np.linalg.inv(temp1)
    temp2 = _beta_helper(pred_std, phylo_std, weights)
    beta = tmp1_inv.dot(temp2)
    if len(beta.shape) == 1:
        beta = beta.reshape((beta.shape[0], 1))  # pragma: no cover
    if use_lock:  # pragma: no cover
        lock.release()
    return beta


# .............................................................................
def _calculate_r_squared(y_hat, phylo_std):
    """Calculates the R-squared value for the inputs.

    Args:
        y_hat (Matrix): A predicted value for a regression
            (n [sites] by k [nodes]).
        phylo_std (Matrix): A standardized phylo matrix
            (n [sites] by k [nodes]).

    Note:
        * R^2 = trace(y_hat . y_hat) / trace(P . P_T)
        * y_hat_T is the transverse of y_hat
        * P_T is the transverse of phylo_std

    Returns:
        Matrix: R squared value for the inputs.
    """
    # t1 = np.trace(y_hat.dot(y_hat.T))
    t_1 = _trace_mtx_by_transverse(y_hat)
    # t2 = np.trace(phylo_std.dot(phylo_std.T))
    t_2 = _trace_mtx_by_transverse(phylo_std)
    r_2 = t_1 / t_2
    # r2 = np.trace(
    #    y_hat.T.dot(y_hat)) / np.trace(phylo_std.T.dot(phylo_std.T))
    return r_2


# .............................................................................
def _calculate_y_hat(pred_std, beta):
    """Calculates the predicted value for a regression (Y hat).

    Args:
        pred_std (Matrix): A standardized predictor matrix
            (n [sites] by i [predictors]).
        beta (Matrix): The regression model associated with this prediction
            (i by k [nodes]).

    Returns:
        Matrix: An (n [sites] by k [nodes]) numpy ndarray
    """
    return pred_std.dot(beta)


# .............................................................................
def _mcpa_for_node(incidence_mtx, env_mtx, bg_mtx, phylo_col, use_locks=False):
    """Runs MCPA computations for a single tree node.

    Args:
        incidence_mtx (Matrix): An incidence matrix (PAM).
        env_mtx (Matrix): An environmental matrix (GRIM).
        bg_mtx (Matrix): A matrix of encoded Biogeographic hypotheses.
        phylo_col (Matrix): A column from the phylo matrix for a
            single node.
        use_locks (boolean): Indicator if locks are needed for larger
            computations.  This is probably only true for parallel runs.

    Returns:
        tuple: Tuple of observed Matrix, f-values Matrix
    """
    species_present_at_node = np.where(phylo_col != 0)[0]
    phylo_col = phylo_col[species_present_at_node, :]

    # Purge incidence matrix to only those species present for this node
    incidence_temp = incidence_mtx[:, species_present_at_node]

    # Get the sites present for this node and purge the other matrices
    sites_present = np.where(incidence_temp.sum(axis=1) > 0)
    incidence = incidence_temp[sites_present]
    env_predictors = env_mtx[sites_present]
    bg_predictors = bg_mtx[sites_present]

    # Get the number of sites and check that all matrices have been purged
    num_sites = sites_present[0].size
    num_env_predictors = env_predictors.shape[1]
    num_bg_predictors = bg_predictors.shape[1]

    obs_values = np.zeros((1, num_env_predictors + num_bg_predictors + 2))
    f_values = np.zeros((1, num_env_predictors + num_bg_predictors + 2))

    try:
        if len(sites_present) > 0:
            # Standardize matrices
            site_weights = np.sum(incidence, axis=1)
            species_weights = np.sum(incidence, axis=0)

            e_std = _standardize_matrix(env_predictors, site_weights)
            b_std = _standardize_matrix(bg_predictors, site_weights)
            all_std = np.concatenate([b_std, e_std], axis=1)
            p_std = _standardize_matrix(phylo_col, species_weights)
            p_sigma_std = np.dot(incidence, p_std)
            # Get Beta, Y(hat), Rho, R-squared, F-pseudo
            beta_env_all = _calculate_beta(
                e_std, site_weights, p_sigma_std, use_lock=use_locks
            )
            y_hat_env_all = _calculate_y_hat(e_std, beta_env_all)
            beta_bg_all = _calculate_beta(
                all_std, site_weights, p_sigma_std, use_lock=use_locks
            )
            y_hat_bg_all = _calculate_y_hat(all_std, beta_bg_all)
            env_r2 = _calculate_r_squared(y_hat_env_all, p_sigma_std)
            bg_r2 = _calculate_r_squared(y_hat_bg_all, p_sigma_std)
            try:
                env_adj_r2 = 1.0 - (
                    (num_sites - 1.0) / (num_sites - num_env_predictors - 1.0)
                ) * (1.0 - env_r2)
            except ZeroDivisionError:  # pragma: no cover
                env_adj_r2 = 0.0

            try:
                bg_adj_r2 = 1.0 - (
                    (num_sites - 1.0) / (num_sites - num_bg_predictors - 1.0)
                ) * (1.0 - bg_r2)
            except ZeroDivisionError:  # pragma: no cover
                bg_adj_r2 = 0.0

            # env_f_pseudo_numerator = np.trace(
            #    y_hat_env_all.T.dot(y_hat_env_all))
            env_f_pseudo_numerator = _trace_mtx_by_transverse(y_hat_env_all.T)
            # bg_f_pseudo_numerator = np.trace(
            #    y_hat_bg_all.T.dot(y_hat_bg_all))
            bg_f_pseudo_numerator = _trace_mtx_by_transverse(y_hat_bg_all.T)

            # Pre-calculate the denominator for the F-pseudo stats
            env_denom_temp = p_sigma_std.T - y_hat_env_all.T
            # env_f_pseudo_denominator = np.trace(
            #    env_denom_temp.T.dot(env_denom_temp))
            env_f_pseudo_denominator = _trace_mtx_by_transverse(env_denom_temp.T)
            bg_denom_temp = p_sigma_std.T - y_hat_bg_all.T
            # bg_f_pseudo_denominator = np.trace(
            #    bg_denom_temp.T.dot(bg_denom_temp))
            bg_f_pseudo_denominator = _trace_mtx_by_transverse(bg_denom_temp.T)

            idx = 0
            # Environment
            # For each environmental predictor, compute the semi-partial
            #    correlation and the F-pseudo value
            for i in range(num_env_predictors):
                wo_predictor = np.delete(e_std, i, axis=1)
                predictor = e_std[:, [i]]

                # Semi-partial correlation
                beta_wo_pred = _calculate_beta(
                    wo_predictor, site_weights, p_sigma_std, use_lock=use_locks
                )
                y_hat_wo_pred = _calculate_y_hat(wo_predictor, beta_wo_pred)
                beta_j_i = _calculate_beta(
                    predictor, site_weights, p_sigma_std, use_lock=False
                )
                r2_j_i = _calculate_r_squared(y_hat_wo_pred, p_sigma_std)
                semi_partial = beta_j_i * np.sqrt(env_r2 - r2_j_i) / np.abs(beta_j_i)
                f_pseudo_env_i = (env_r2 - r2_j_i) / env_f_pseudo_denominator
                obs_values[0, idx] = semi_partial
                f_values[0, idx] = f_pseudo_env_i
                idx += 1
            # Add Environment adjusted R squared
            obs_values[0, idx] = env_adj_r2
            f_values[0, idx] = env_f_pseudo_numerator / env_f_pseudo_denominator
            idx += 1

            # Biogeography
            for i in range(num_bg_predictors):
                wo_predictor = np.delete(all_std, i, axis=1)
                predictor = all_std[:, [i]]

                # Semi-partial correlation
                beta_wo_pred = _calculate_beta(
                    wo_predictor, site_weights, p_sigma_std, use_lock=use_locks
                )
                y_hat_wo_pred = _calculate_y_hat(wo_predictor, beta_wo_pred)
                beta_j_i = _calculate_beta(
                    predictor, site_weights, p_sigma_std, use_lock=False
                )
                r2_j_i = _calculate_r_squared(y_hat_wo_pred, p_sigma_std)
                semi_partial = beta_j_i * np.sqrt(bg_r2 - r2_j_i) / np.abs(beta_j_i)
                f_pseudo_bg_i = (bg_r2 - r2_j_i) / bg_f_pseudo_denominator
                obs_values[0, idx] = semi_partial
                f_values[0, idx] = f_pseudo_bg_i
                idx += 1
            # Add Biogeography adjusted R squared
            obs_values[0, idx] = bg_adj_r2
            f_values[0, idx] = bg_f_pseudo_numerator / bg_f_pseudo_denominator
    except np.linalg.linalg.LinAlgError:
        # Singular matrix that does not have inverse
        pass

    return (obs_values, f_values)


# .............................................................................
def _standardize_matrix(mtx, weights):
    """Standardizes a phylogenetic or predictor matrix.

    Args:
        mtx (Matrix): The matrix to standardize
        weights (Matrix): A one-dimensional array of sums to use for standardization.

    Note:
        * Formula for standardization ::
            Mstd = M - 1c.1r.W.M(1/trace(W)) ./ 1c(1r.W(M*M)
                       - ((1r.W.M)*(1r.W.M)(1/trace(W))(1/trace(W)-1))^0.5
        * M - Matrix to be standardized
        * W - A k by k diagonal matrix of weights, where each non-zero value is
            the column or row sum (depending on the M) for an incidence matrix.
        * 1r - A row of k ones
        * 1c - A column of k ones
        * trace - Returns the sum of the input matrix
        * "./" indicates Hadamard division
        * "*" indicates Hadamard multiplication
        * Code adopted from supplemental material MATLAB code

    See:
        Literature supplemental materials

    Returns:
        Matrix: Standardized matrix.
    """
    # Create a row of ones, we'll transpose for a column
    ones = np.ones((1, weights.shape[0]), dtype=float)
    # This maps to trace(W)
    total_sum = np.sum(weights)

    # s1 = 1r.W.M
    s_1 = (ones * weights).dot(mtx)
    # s2 = 1r.W.(M*M)
    s_2 = (ones * weights).dot(mtx * mtx)

    mean_weighted = s_1 / total_sum
    std_dev_weighted = ((s_2 - (s_1**2.0 / total_sum)) / (total_sum)) ** 0.5

    # Fixes any invalid values created previously
    tmp = np.nan_to_num(ones.T.dot(std_dev_weighted) ** -1.0)
    std_mtx = tmp * (mtx - ones.T.dot(mean_weighted))

    return std_mtx


# .............................................................................
def _trace_mtx_by_transverse(mtx):
    """Prevent a memory bomb by performing trace(mtx . mtx.T) in a smarter way.

    This method takes advantage of the fact that we are really only interested
        in the diagonal matrix created by performing a dot product of a matrix
        and its transverse.  Because of that, we can perform the dot product of
        each row by its transverse and then sum the results.

    Args:
        mtx (Matrix): A matrix to use to calculate the trace(M . M_T).

    Note:
        * There was an error in the formula that called for a dot product of
            each matrix transverse with itself.  This was a typo as that would
            only work with square matrices.

    Returns:
        Matrix: Trace matrix dot matrix transverse.
    """
    return np.sum([row.dot(row.T) for row in mtx])


# .............................................................................
def get_p_values(observed_value, test_values, num_permutations=None):
    """Gets an array of P-Values.

    Gets an (1 or 2 dimension) array of P values where the P value for an array
        location is determined by finding the number of test values at
        corresponding locations are greater than or equal to that value and
        then dividing that number by the number of permutations

    Args:
        observed_value (Matrix): An array of observed values to use as a
            reference.
        test_values (Matrix): A list of arrays generated from randomizations
            that will be compared to the observed
        num_permutations: (optional) The total number of randomizations
            performed.  Divide the P-values by this if provided.

    Todo:
        Deprecate this in favor of method in running stats

    Returns:
        Matrix: Calculated p-values.
    """
    if num_permutations is None:
        num_permutations = 1.0

    # Create the P-Values matrix
    p_vals = np.zeros(observed_value.shape, dtype=float)
    # For each matrix in test values
    for test_mtx in test_values:
        # Add 1 where every value in the test matrix is greater than or equal
        #    to the value in the observed value.  Numpy comparisons will create
        #    a matrix of boolean values for each cell, which when added to the
        #    p_vals matrix will be treated as 1 for True and 0 for False

        # If this is a stack
        if test_mtx.ndim == 3:
            for i in range(test_mtx.shape[2]):
                p_vals += np.abs(np.round(test_mtx[:, :, i], 5)) >= np.abs(
                    np.round(observed_value, 5)
                )
        else:
            p_vals += np.abs(np.round(test_mtx, 5)) >= np.abs(
                np.round(observed_value, 5)
            )
    # Reshape and adding depth header
    if len(p_vals.shape) == 2:
        p_vals = np.expand_dims(p_vals, axis=2)
    headers = observed_value.headers
    headers['2'] = ['P-values']
    # Scale and return the p-values matrix
    ret_data_tmp = np.nan_to_num(p_vals / num_permutations)
    return Matrix(np.clip(ret_data_tmp, -1.0, 1.0), headers=headers)


# .............................................................................
def mcpa(incidence_matrix, phylo_mtx, env_mtx, bg_mtx):
    """Runs MCPA for a set of matrices.

    Args:
        incidence_matrix (Matrix): A binary Lifemapper Matrix object
            representing the incidence of each species for each site by coding
            them as ones.  This is the same thing as a Lifemapper Presence
            Absence Matrix, or PAM (n [sites] by k+1 [species]).
        phylo_mtx (Matrix): A matrix encoding of a phylogenetic tree where each
            cell represents the relative contribution of each tip to each
            inner tree node (k+1 [species] by k [nodes]).
        env_mtx (Matrix): A matrix encoding of the environment for each site
            (n [sites] by ei [environmental predictors]).
        bg_mtx (Matirx): A matrix of Helmert contrasts (-1, 0, 1) for
            Biogeographic hypotheses (n [sites] by bi [biogeographic
            predictors]).

    Returns:
        tuple: Tuple of Matrix of observed values and Matrix of F-pseudo values.
    """
    site_present = np.any(incidence_matrix, axis=1)
    empty_sites = np.where(site_present == 0)[0]

    # Initial purge of empty sites
    init_incidence = np.delete(incidence_matrix, empty_sites, axis=0)
    env_predictors = np.delete(env_mtx, empty_sites, axis=0)
    bg_predictors = np.delete(bg_mtx, empty_sites, axis=0)

    num_nodes = phylo_mtx.shape[1]
    num_predictors = env_predictors.shape[1] + bg_predictors.shape[1]

    obs_results = np.empty((num_nodes, num_predictors + 2))
    f_results = np.empty((num_nodes, num_predictors + 2))
    for i in range(num_nodes):
        # print('Node {} of {}'.format(i+1, num_nodes))
        obs, f_vals = _mcpa_for_node(
            init_incidence, env_predictors, bg_predictors, phylo_mtx[:, [i]]
        )
        obs_results[i] = obs
        f_results[i] = f_vals

    print('Processed mcpa for {} of {} nodes'.format(i + 1, num_nodes))

    # Correct any nans and add depth
    obs_results = np.clip(np.expand_dims(np.nan_to_num(obs_results), axis=2), -1.0, 1.0)
    f_results = np.clip(np.expand_dims(np.nan_to_num(f_results), axis=2), -1.0, 1.0)

    column_headers = env_mtx.get_column_headers()
    column_headers.append('Env - Adjusted R-squared')
    column_headers.extend(bg_mtx.get_column_headers())
    column_headers.append('BG - Adjusted R-squared')
    obs_headers = {
        '0': phylo_mtx.get_column_headers(),
        '1': column_headers,
        '2': ['Observed'],
    }
    f_headers = {
        '0': phylo_mtx.get_column_headers(),
        '1': column_headers,
        '2': ['F-values'],
    }
    obs_mtx = Matrix(obs_results, headers=obs_headers)
    f_mtx = Matrix(f_results, headers=f_headers)
    return (obs_mtx, f_mtx)


# .............................................................................
def mcpa_parallel(incidence_matrix, phylo_mtx, env_mtx, bg_mtx):  # pragma: no cover
    """Run MCPA for a set of matrices using parallelism.

    Performs MCPA across each of the tree nodes in parallel.

    Args:
        incidence_matrix (Matrix): A binary Lifemapper Matrix object
            representing the incidence of each species for each site by coding
            them as ones.  This is the same thing as a Lifemapper Presence
            Absence Matrix, or PAM (n [sites] by k+1 [species]).
        phylo_mtx (Matrix): A matrix encoding of a phylogenetic tree where each
            cell represents the relative contribution of each tip to each
            inner tree node (k+1 [species] by k [nodes]).
        env_mtx (Matrix): A matrix encoding of the environment for each site
            (n [sites] by ei [environmental predictors]).
        bg_mtx (Matirx): A matrix of Helmert contrasts (-1, 0, 1) for
            Biogeographic hypotheses (n [sites] by bi [biogeographic
            predictors]).

    Returns:
        tuple: Tuple of Matrix of observed values and Matrix of F-pseudo values.
    """
    site_present = np.any(incidence_matrix, axis=1)
    empty_sites = np.where(site_present == 0)[0]

    # Initial purge of empty sites
    init_incidence = np.delete(incidence_matrix, empty_sites, axis=0)
    env_predictors = np.delete(env_mtx, empty_sites, axis=0)
    bg_predictors = np.delete(bg_mtx, empty_sites, axis=0)

    num_nodes = phylo_mtx.shape[1]
    num_predictors = env_predictors.shape[1] + bg_predictors.shape[1]

    obs_results = np.empty((num_nodes, num_predictors + 2))
    f_results = np.empty((num_nodes, num_predictors + 2))

    func = partial(
        _mcpa_for_node, init_incidence, env_predictors, bg_predictors, use_locks=True
    )

    # Use an Executor to parallelize the computations over each tree node
    # Note: The executor class is determined at the module level, so see top of
    #    module for more information about executor class and concurrency
    with ExecutorClass(CONCURRENCY_FACTOR) as executor:
        comp_run = executor.map(func, [phylo_mtx[:, [i]] for i in range(num_nodes)])
        i = 0
        for obs, f_val in comp_run:
            obs_results[i] = obs
            f_results[i] = f_val

    # Correct any nans
    obs_results = np.clip(np.nan_to_num(obs_results), -1.0, 1.0)
    f_results = np.clip(np.nan_to_num(f_results), -1.0, 1.0)

    column_headers = env_mtx.get_column_headers()
    column_headers.append('Env - Adjusted R-squared')
    column_headers.extend(bg_mtx.get_column_headers())
    column_headers.append('BG - Adjusted R-squared')
    obs_headers = {
        '0': phylo_mtx.get_column_headers(),
        '1': column_headers,
        '2': ['Observed'],
    }
    f_headers = {
        '0': phylo_mtx.get_column_headers(),
        '1': column_headers,
        '2': ['F-values'],
    }
    obs_mtx = Matrix(obs_results, headers=obs_headers)
    f_mtx = Matrix(f_results, headers=f_headers)
    return (obs_mtx, f_mtx)



"""Module that contains a Matrix class that has header information.

Note:
    * Subclassing based on
        https://docs.scipy.org/doc/numpy/user/basics.subclassing.html

Todo:
    * Handle multiple rows / columns / etc of headers like:
        (PAM x, y, site ids).
    * Load should handle compressed and not compressed.
    * Watch NEP-0018 for Numpy function overrides (ex. concatenate)
"""
from collections.abc import KeysView
from copy import deepcopy
import io
import json
import zipfile

import numpy as np


HEADERS_KEY = 'headers'
METADATA_KEY = 'metadata'
VERSION_KEY = 'version'
VERSION = '2.0.2'
HEADERS_FILENAME = 'headers.json'
DATA_FILENAME = 'data.npz'


# .............................................................................
class Matrix(np.ndarray):
    """Lifemapper wrapper for Numpy ndarrays that adds headers."""

    # ...........................
    def __new__(cls, input_array, headers=None, metadata=None):
        """Create a new Matrix object from an existing ndarray.

        Args:
            input_array (numpy.ndarray): An existing ndarray
            headers (:obj:`dict` or :obj:`list` of :obj:`list` of :obj:`str`):
                Optional headers for this matrix.  This may be either a list of
                lists, where the index of a list in the lists will be treated
                as the axis::
                    (ex. [['Row 1', 'Row 2', 'Row 3'],
                          ['Column 1', 'Column 2']])
                Or this could be a dictionary where the key is used for the
                axis.  (Ex:
                    {
                        '1' : ['Column 1', 'Column 2'],
                        '0' : ['Row 1', 'Row 2', 'Row 3']
                    }
            metadata (:obj:`dict`): Optional metadata about his matrix.

        Note:
            * Triggers a call to Matrix.__array_finalize__.

        Returns:
            Matrix: A converted numpy array to Matrix.
        """
        # input_array is already formed ndarray instance, cast to our type
        obj = np.asarray(input_array).view(cls)

        # Set the 'headers' attribute to the value passed
        if headers is None:
            headers = {}
        if metadata is None:
            metadata = {}
        obj.headers = headers
        obj.metadata = metadata
        return obj

    # ...........................
    def __array_finalize__(self, obj):
        """Overridden function from ndarray.

        ``self`` is a new object resulting from ndarray.__new__(Matrix, ...),
        therefore it only has the attributes that the ndarray.__new__
        constructor gave it.

        Args:
            obj (:obj:`Matrix1): If the object is a matrix, pull out headers and
                metadata.
        """
        # We could have got to the ndarray.__new__ call in 3 ways:
        # 1. From an explicit constructor - Matrix():
        #    obj is None
        #    We're in the middle of the Matrix.__new__ constructor, and
        #        self.headers will be set when we return to Matrix.__new__
        if obj is None:  # pragma: no cover
            return

        # 2. From casting - arr.view(Matrix):
        #    obj is arr
        #    type(obj) can be Matrix
        # 3. From new-from-template - mtx[:3]
        #    type(obj) is Matrix

        # Note: It is is here, rather than the __new__ method, that we set the
        #    default value for 'headers', because this method sees all creation
        #    of default objects
        self.headers = getattr(obj, 'headers', {})
        self.metadata = getattr(obj, 'metadata', {})

    # ...........................
    @classmethod
    def load(cls, filename):
        """Load a matrix from a filename.

        Args:
            filename (:obj:`str`): File location of matrix to load.

        Returns:
            Matrix: The matrix read from the file.
        """
        with open(filename, 'rb') as in_file:
            return Matrix.load_flo(in_file)

    # ...........................
    @classmethod
    def load_csv(cls, flo, dtype=float, num_header_rows=0, num_header_cols=0):
        """Attempts to load a Matrix object from a CSV file-like object.

        Args:
            flo (File-like object): A file like object containing csv data.
            dtype (:obj:`method`, optional): The data type for the data.  Will
                be used to cast data when adding to matrix.
            num_header_rows (:obj:`int`, optional): The number of header rows
                in the CSV file.
            num_header_cols (:obj:`int`, optional): The number of header
                columns in the CSV file.

        Returns:
            Matrix: The newly loaded Matrix object.
        """
        col_headers = []
        row_headers = []
        header_lines = []  # Leading rows that are headers
        data = []
        i = 0
        for line in flo:
            items = line.strip().split(',')
            # If header row, add to header rows for processing
            if i < num_header_rows:
                # Add the headers to header lines for processing
                header_lines.append(items[num_header_cols:])
            else:
                if num_header_cols == 1:
                    row_headers.append(items[0].strip())
                elif num_header_cols > 1:
                    row_headers.append([q.strip() for q in items[:num_header_cols]])
                data.append([dtype(x) for x in items[num_header_cols:]])

            i += 1

        # Process header columns from header rows
        if num_header_rows == 1:
            col_headers = [q.strip() for q in header_lines[0]]
        elif num_header_rows > 1:
            for j in range(len(header_lines[0])):
                hdr = []
                for hdr_idx in range(num_header_rows):
                    hdr.append(header_lines[hdr_idx][j].strip())
                col_headers.append(hdr)

        data_array = np.array(data)

        return cls(data_array, headers={'0': row_headers, '1': col_headers})

    # ...........................
    @classmethod
    def load_flo(cls, flo):
        """Attempts to load a Matrix object from a file.

        Args:
            flo (file-like): A file-like object with matrix data.

        Returns:
            Matrix: The newly loaded Matrix object.
        """
        with zipfile.ZipFile(flo) as zip_f:
            my_obj = json.loads(zip_f.read(HEADERS_FILENAME).decode('utf-8'))
            data_bytes = io.BytesIO()
            data_bytes.write(zip_f.read(DATA_FILENAME))
            data_bytes.seek(0)
            tmp = np.load(data_bytes)
            data = tmp[tmp.files[0]]
            # data = np.array(tmp[list(tmp.keys())[0]])
            data_bytes.close()

        return cls(data, headers=my_obj[HEADERS_KEY])

    # ...........................
    @classmethod
    def concatenate(cls, mtx_list, axis=0):
        """Concatenates multiple Matrix objects together to form a new Matrix.

        Args:
            mtx_list (:obj:`list` of :obj:`Matrix`): A List of Matrix objects
                to concatenate together.
            axis (:obj:`int`, optional): The axis to concatenate these Matrix objects
                on.  This should be an integer for the matrix axis of concatenation.
                This will be converted to a string where needed for headers.

        Note:
            * Assumes that headers for other axes are the same.

        Returns:
            Matrix: The concatenated Matrix objects.
        """
        mtx_objs = []
        axis_headers = []
        first_mtx = None
        for mtx in mtx_list:
            # Make sure we reshape if necessary if adding new axis (stacking)
            if mtx.ndim < axis + 1:  # Add 1 since zero-based
                mtx = Matrix(np.expand_dims(mtx, axis), headers=mtx.headers)
                mtx.set_headers([''], axis=str(axis))
            # Cast mtx to Matrix in case it is not
            mtx = mtx.view(Matrix)
            h = mtx.get_headers(axis=str(axis))
            if h is None:
                h = ['']
            axis_headers.extend(h)
            mtx_objs.append(mtx)
            if first_mtx is None:
                first_mtx = mtx
        # Create a new Matrix, use the first matrix for base headers and
        #    metadata
        new_mtx = cls(
            np.concatenate(mtx_objs, axis=axis),
            headers=first_mtx.get_headers(),
            metadata=first_mtx.get_metadata(),
        )
        # Set the headers for the new axis
        new_mtx.set_headers(axis_headers, axis=str(axis))
        return new_mtx

    # ...........................
    def flatten_2d(self):
        """Flattens a higher dimension Matrix object into a 2D matrix.

        Todo:
            * Modify this method to take an argument for the number of
                dimensions that the matrix should be flattened to.

        Returns:
            Matrix: A Matrix object flattened to only have two dimensions.
        """
        flat_mtx = self
        while flat_mtx.ndim > 2:
            # More than two dimensions so we must flatten
            old_shape = flat_mtx.shape
            old_num_rows = old_shape[0]
            new_shape = tuple(
                [old_shape[0] * old_shape[2], old_shape[1]] + list(old_shape[3:])
            )
            new_mtx = Matrix(np.zeros(new_shape))

            old_rh = flat_mtx.get_row_headers()
            if old_rh is None:
                old_rh = []
            new_rh = []

            # Get old headers
            old_headers = flat_mtx.get_headers(axis=2)
            if old_headers is None:
                old_headers = [''] * old_shape[2]

            # Set data and headers
            for i in range(old_shape[2]):
                oh = old_headers[i]
                # Set data
                start_row = i * old_num_rows
                end_row = (i + 1) * old_num_rows
                new_mtx[start_row:end_row, :] = flat_mtx[:, :, i]

                # Set row headers
                for rh in old_rh:
                    if not isinstance(rh, list):
                        rh = [rh]
                    new_rh.append(rh + [oh])

            # Set the headers on the new matrix
            new_mtx.set_row_headers(new_rh)
            new_mtx.set_column_headers(flat_mtx.get_column_headers())

            # Higher order headers
            for axis in flat_mtx.headers.keys():
                if int(axis) > 2:
                    # Reduce the key of the axis by one and set headers on
                    #   new matrix
                    new_mtx.set_headers(
                        flat_mtx.get_headers(axis=axis), axis=str(int(axis) - 1)
                    )

            flat_mtx = new_mtx

        return flat_mtx

    # ...........................
    def get_column_headers(self):
        """Shortcut to get column headers.

        Returns:
            A list of headers for each column.
        """
        return self.get_headers(axis=1)

    # ...........................
    def get_headers(self, axis=None):
        """Gets the headers associated with this Matrix.

        Args:
            axis (:obj:`int`, optional): If provided, return headers for this
                axis, else, return all.

        Returns:
            dict: If axis is None, a dictionary of all headers for the matrix.
            list: If axis is int, A list of headers for the specified axis.
        """
        try:
            if axis is None:
                return self.headers

            if str(axis) in self.headers.keys():
                return self.headers[str(axis)]

            return None
        except Exception:  # pragma: no cover
            return {}

    # ...........................
    def get_metadata(self):
        """Retrieves matrix metadata.

        Returns:
            dict: A dictionary of metadata for the matrix.
        """
        return self.metadata

    # ...........................
    def get_row_headers(self):
        """Shortcut to get row headers.

        Returns:
            A list of headers for the rows in the matrix.
        """
        return self.get_headers(axis=0)

    # ...........................
    def save(self, flo):
        """Saves the Matrix to a file-like object.

        Saves the Matrix object in a JSON / Numpy zip file to the file-like
        object.

        Args:
            flo (file-like): The file-like object to write to.
        """
        my_obj = {}
        my_obj[HEADERS_KEY] = self.headers
        my_obj[METADATA_KEY] = self.metadata
        my_obj[VERSION_KEY] = VERSION

        np_bytes = io.BytesIO()
        np.savez_compressed(np_bytes, self)
        np_bytes.seek(0)
        zip_np_str = np_bytes.getvalue()
        np_bytes.close()

        with zipfile.ZipFile(
            flo, mode='w', compression=zipfile.ZIP_DEFLATED, allowZip64=True
        ) as zip_f:

            zip_f.writestr(HEADERS_FILENAME, json.dumps(my_obj, default=float))
            zip_f.writestr(DATA_FILENAME, zip_np_str)

    # ...........................
    def set_column_headers(self, headers):
        """Shortcut to set column headers.

        Args:
            headers (:obj:`list` of :obj:`str`): A list of new column headers.
        """
        self.set_headers(headers, axis=1)

    # ...........................
    def set_headers(self, headers, axis=None):
        """Sets the headers for this Matrix.

        Args:
            headers (:obj:`dict` or :obj:`list` of :obj:`list` or :obj:`list`): Matrix
                headers.  Can be a list of lists, a dictionary of lists, or if axis is
                provided, a single list.
            axis (:obj:`int`): If provided, set the headers for a specific axis, else,
                process  as if it is for the entire Matrix.

        Todo:
            Validate input for single axis operation?

        Note:
            Resets headers dictionary when setting values for all headers.
            Duck types to use list of lists or dictionary to set values for different
                axes.
        """
        if isinstance(headers, KeysView):
            headers = list(headers)

        if axis is not None:
            self.headers[str(axis)] = headers
        else:
            self.headers = {}

            for k in headers.keys():
                self.headers[str(k)] = headers[str(k)]

    # ...........................
    def set_row_headers(self, headers):
        """Shortcut to set row headers.

        Args:
            headers (:obj:`list` of :obj:`str`): A list of new row headers.
        """
        self.set_headers(headers, axis=0)

    # ...........................
    def slice(self, *args):
        """Subsets the matrix and returns a new instance.

        Args:
            *args: A variable length argument list of iterables for the indices
                to retrieve for each axis.

        Note:
            * The first parameter will be for axis 0, second for axis 1, etc.

        Returns:
            Matrix: A new Matrix that is a subset of the original specified by
                the slicing parameters.
        """
        new_data = np.copy(self)
        new_headers = deepcopy(self.headers)
        # For each arg in the list
        for i in range(len(args)):
            # Subset the data matrix
            new_data = new_data.take(args[i], axis=i)
            # Subset the headers
            tmp = []
            if str(i) in new_headers.keys():
                for j in args[i]:
                    tmp.append(new_headers[str(i)][j])
            new_headers[str(i)] = tmp
        return Matrix(new_data, headers=new_headers)

    # ...........................
    def slice_by_header(self, header, axis):
        """Gets a slice of the Matrix matching the header provided.

        Args:
            header (:obj:`str`): The name of a header to use for slicing
            axis (:obj:`int`): The axis to find this header.

        Todo:
            Add capability to slice over multiple axes and multiple headers.
                Maybe combine with other slice method and provide method to
                search for header indices.

        Returns:
            Matrix: A subset of the original Matrix specified by the header.
        """
        idx = self.headers[str(axis)].index(header)
        new_data = np.copy(np.take(self, idx, axis=axis))

        # Need to reshape the result.  Take the existing shape and change the
        #    query axis to 1
        new_shape = list(self.shape)
        new_shape[axis] = 1
        new_data = new_data.reshape(new_shape)

        # Copy the headers and set the header for the axis to just be the
        #    search header
        new_headers = deepcopy(self.headers)
        new_headers[str(axis)] = [header]

        # Return a new Matrix
        return Matrix(new_data, headers=new_headers, metadata=self.metadata)

    # ...........................
    @property
    def T(self):
        """Get the transpose of the matrix.

        Returns:
            Matrix: The matrix transpose.
        """
        if self.ndim < 2:
            return self

        return self.transpose()

    # ...........................
    def transpose(self, *axes):
        """Transposes the Matrix.

        Args:
            axes (:obj:`None`, :obj:`tuple` of :obj:`int`, n :obj:`int`s): The order of
                the axes in the transposition. see: ndarray.transpose

        Returns:
            Matrix: A transposed version of the original matrix
        """
        new_mtx = Matrix(self.view(np.ndarray).transpose(*axes))
        if len(axes) == 0:
            dim_order = range(len(self.shape) - 1, -1, -1)
        elif isinstance(axes[0], tuple):
            dim_order = list(axes[0])
        else:
            dim_order = list(axes)
        # Set headers
        for i, old_dim in enumerate(dim_order):
            old_dim = str(old_dim)
            try:
                if old_dim in self.get_headers().keys():
                    new_mtx.set_headers(self.get_headers(axis=old_dim), axis=str(i))
            except Exception as err:  # pragma: no cover
                print(err)
                print(dir(self))
        return new_mtx

    # ...........................
    def write(self, filename):
        """Write the matrix to the specified file location.

        Args:
            filename (:obj:`str`): The file location to save to.
        """
        with open(filename, 'wb') as out_file:
            self.save(out_file)

    # ...........................
    def write_csv(self, flo, *slice_args):
        """Writes the Matrix object to a CSV file-like object.

        Args:
            flo (file-like): The file-like object to write to.
            *slice_args: A variable length argument list of iterables to use
                for a slice operation prior to generating CSV content.

        Todo:
            Handle header overlap (where the header for one axis is for another
                axis header.

        Note:
            Currently only works for 2-D tables.
        """
        if list(slice_args):
            mtx = self.slice(*slice_args)
        else:
            mtx = self

        if mtx.ndim > 2:
            mtx = mtx.flatten_2d()

        # .....................
        # Inner functions

        # .....................
        def already_lists(x):
            """Use this function for processing headers when they are lists.

            Args:
                x (:obj:`list`): A list value to return.

            Returns:
                list: A list of data.
            """
            return x

        # .....................
        def make_lists(x):
            """Use this function for processing non-list headers.

            Args:
                x (:obj:`object`): A non-list value to modify.

            Returns:
                list: A list of data.
            """
            return [x]

        # .....................
        def csv_generator():
            """Generator that yields rows of values to be output as CSV.

            Yields:
                list: A list of data for a row.
            """
            try:
                row_headers = mtx.headers['0']
            except (KeyError, TypeError):
                # No row headers
                row_headers = [[] for _ in range(mtx.shape[0])]

            if isinstance(row_headers[0], list):
                listify = already_lists
            else:
                listify = make_lists

            # Start with the header row, if we have one
            if '1' in mtx.headers and mtx.headers['1']:
                # Make column headers lists of lists
                if not isinstance(mtx.headers['1'][0], (tuple, list)):
                    header_row = [''] * len(
                        listify(row_headers[0]) if row_headers else []
                    )
                    header_row.extend(mtx.headers['1'])
                    yield header_row
                else:
                    for i in range(len(mtx.headers['1'][0])):
                        header_row = [''] * len(
                            listify(row_headers[0]) if row_headers else []
                        )
                        header_row.extend(
                            [
                                mtx.headers['1'][j][i]
                                for j in range(len(mtx.headers['1']))
                            ]
                        )
                        yield header_row
            # For each row in the data set
            for i in range(mtx.shape[0]):
                # Add the row headers if exists
                row = []
                row.extend(listify(row_headers[i]))
                # Get the data from the data array
                row.extend(mtx[i].tolist())
                yield row

        # .....................
        # Main write_csv function
        for row in csv_generator():
            flo.write(u"{}\n".format(','.join([str(v) for v in row])))


# .............................................................................
# Set the module for Matrix to be lmpy
Matrix.__module__ = 'lmpy'

__all__ = ['Matrix']



"""Module containing common format occurrence data wrangler."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler
from lmpy.point import Point


# .....................................................................................
class CommonFormatWrangler(_OccurrenceDataWrangler):
    """Modifies a point record using an attribute map."""
    name = 'CommonFormatWrangler'
    version = '1.0'

    # .......................
    def __init__(self, attribute_map, **params):
        """Constructor for CommonFormatWrangler class.

        Args:
            attribute_map (dict): A mapping of source key, target values.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.attribute_map = attribute_map
        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _modify_point(self, point):
        """Update point attributes.

        Args:
            point (Point): A point to modify.

        Returns:
            Point, bool: Modified (if needed) point and boolean if point was modified.
        """
        new_pt = Point(
            point.species_name,
            point.x,
            point.y,
            attributes={
                new_name: point.get_attribute(old_name)
                for old_name, new_name in self.attribute_map.items()
            },
        )

        return new_pt, True



"""Module containing base class for data wranglers."""


# .....................................................................................
class _DataWrangler:
    """Data wrangler base class."""
    name = '_DataWrangler'
    version = '1.0'

    # ........................
    def __init__(self, *args, **kwargs):
        """Base class constructor.

        Args:
            *args (tuple): Positional arguments.
            **kwargs (dict): Dictionary of parameters.
        """
        self.report = {}

    # ........................
    @classmethod
    def from_config(cls, config):
        """Get an instance generated from the provided configuration.

        Args:
            config (dict): A dictionary of configuration parameters.

        Returns:
            _DataWrangler: The instance created from the parameters.
        """
        return cls(**config)

    # ........................
    def get_report(self):
        """Get a report of the results of the wrangler.

        Returns:
            dict: A dictionary of reporting information.
        """
        self.report['name'] = self.name
        self.report['version'] = self.version
        return self.report



"""Matrix data wranglers."""

from . import accepted_name_wrangler
from . import base
from . import match_tree_wrangler
from . import purge_empty_slices_wrangler
from . import subset_reorder_slices_wrangler

__all__ = [
    'accepted_name_wrangler',
    'base',
    'match_tree_wrangler',
    'purge_empty_slices_wrangler',
    'subset_reorder_slices_wrangler',
]



"""Convert a lmpy Matrix to a (.csv) file."""
import argparse

from lmpy import Matrix


DESCRIPTION = 'Convert a lmpy Matrix to a CSV file with numerical values.'


# .....................................................................................
def convert_lmm_to_csv(mtx, csv_filename):
    """Convert a lmpy Matrix to a csv file.

    Args:
        mtx (Matrix): A lmpy matrix to convert to csv.
        csv_filename (str): The file location of the csv to convert.
    """
    with open(csv_filename, mode='wt') as csv_out:
        mtx.write_csv(csv_out)


# .....................................................................................
def cli():
    """Provide a command-line tool for converting lmms to csvs."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        'in_lmm_filename', type=str, help='Lmpy LMM filename to convert to CSV.'
    )
    parser.add_argument(
        'out_csv_filename', type=str, help='Location to write the converted matrix CSV.'
    )
    args = parser.parse_args()
    mtx = Matrix.load(args.in_lmm_filename)
    convert_lmm_to_csv(mtx, args.out_csv_filename)


# .....................................................................................
__all__ = ['cli', 'convert_lmm_to_csv']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Data wrangler factory."""
import importlib
import inspect

from lmpy.data_wrangling.base import _DataWrangler


# .....................................................................................
class WranglerFactory:
    """Factory for configuring data wranglers."""

    # .......................
    def __init__(self):
        """Constructor for the WranglerFactory class."""
        self.wrangler_types = {}
        self._load_wranglers()

    # .......................
    def _load_wranglers(self):
        """Load possible data wrangler types."""

        def _find_data_wranglers(base_module, base_class):
            """Recursive function for finding data wranglers for factory.

            Args:
                base_module (Module): A python module to traverse.
                base_class (Class): Returned classes should be subclasses of this.
            """
            for element_name in dir(base_module):
                # Ignore anything that starts with an underscore
                if not element_name.startswith("_"):
                    # Check the element
                    element = getattr(base_module, element_name)

                    if inspect.isclass(element) and issubclass(element, base_class):
                        # If we found a subclass of our base class, add it
                        self.wrangler_types[element.name] = element
                    elif inspect.ismodule(element) and element.__package__.startswith(
                        base_module.__package__
                    ):
                        # If this is a submodule of the provided module, recurse
                        _find_data_wranglers(element, base_class)

        # Load data_wrangling package now to avoid circular imports
        _find_data_wranglers(
            importlib.import_module("lmpy.data_wrangling"), _DataWrangler
        )

    # .......................
    def get_wranglers(self, wrangler_configs):
        """Get configured data wranglers.

        Args:
            wrangler_configs (list of dicts): A list of wrangler configuration
                dictionaries.

        Returns:
            list of _DataWrangler subclasses: A List of configured data wranglers.
        """
        if isinstance(wrangler_configs, dict):
            wrangler_configs = [wrangler_configs]

        return [
            self.wrangler_types[config["wrangler_type"]].from_config(
                config
            )
            for config in wrangler_configs
        ]



"""This tool cleans occurrence records according to the wranglers specified."""
import argparse
import json

from lmpy.data_wrangling.factory import WranglerFactory
from lmpy.point import PointCsvReader, PointCsvWriter


# .....................................................................................
def clean_data(reader, writer_filename, wranglers, write_fields=None, log_output=False):
    """Clean occurrence data.

    Args:
        reader (PointCsvReader): A reader object that generates Point objects.
        writer_filename (str): A file location to write cleaned points.
        wranglers (list of Wranglers): A list of data wranglers that manipulate
            and / or filter Points for cleaning.
        write_fields (list or None): A list of Point attributes to write to output CSV.
            If None, determine from first cleaned Point.
        log_output (bool): Should output be logged to console.

    Returns:
        dict: Output report from data wrangling.
    """
    if log_output:

        def log_msg(msg):
            print(msg)

    else:

        def log_msg(msg):
            pass

    report = {
        'input_records': 0,
        'output_records': 0,
        'wranglers': [wrangler.get_report() for wrangler in wranglers],
    }
    # Open reader
    reader.open()
    writer = None
    for points in reader:
        report['input_records'] += len(points)
        for wrangler in wranglers:
            wrangler_name = wrangler.name
            # If there are points, wrangle them
            if points:
                tmp = len(points)
                sp_name = points[0].species_name
                points = wrangler.wrangle_points(points)
                log_msg(f'{wrangler_name} removed {tmp - len(points)} {sp_name} points')
        # If any points are left, write them
        if points:
            report['output_records'] += len(points)
            if writer is None:
                if write_fields is None:
                    write_fields = points[0].get_attribute_names()
                writer = PointCsvWriter(writer_filename, write_fields)
                writer.open()
            writer.write_points(points)
    # Close reader and writer
    reader.close()
    if writer:
        writer.close()
    return report


# .....................................................................................
def cli():
    """A command-line interface to the tool."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-sp',
        '--species_key',
        type=str,
        default='species_name',
        help='The CSV column header for species name.',
    )
    parser.add_argument(
        '-x',
        '--x_key',
        type=str,
        default='x',
        help='The CSV column header for the X (longitude) field.',
    )
    parser.add_argument(
        '-y',
        '--y_key',
        type=str,
        default='y',
        help='The CSV column header for the Y (latitude) field.',
    )
    parser.add_argument(
        '-r',
        '--report_filename',
        type=str,
        help='File location to write optional output report JSON.',
    )
    parser.add_argument(
        '-l',
        '--log_output',
        action='store_true',
        default=False,
        help='Should output messages be written to console.',
    )
    parser.add_argument(
        'reader_filename', type=str, help='Input occurrence CSV filename.'
    )
    parser.add_argument(
        'writer_filename', type=str, help='Output cleaned occurrence CSV filename.'
    )
    parser.add_argument(
        'wrangler_config_filename', type=str, help='Wrangler(s) configuration file.'
    )
    args = parser.parse_args()

    # Get wranglers
    wrangler_factory = WranglerFactory()
    wranglers = wrangler_factory.get_wranglers(
        json.load(open(args.wrangler_config_filename, mode='rt'))
    )

    # Get reader
    reader = PointCsvReader(
        args.reader_filename, args.species_key, args.x_key, args.y_key
    )

    # Clean data
    report = clean_data(
        reader, args.writer_filename, wranglers, log_output=args.log_output
    )

    # If the output report was requested, write it
    if args.report_filename:
        with open(args.report_filename, mode='wt') as out_file:
            json.dump(report, out_file, indent=4)


# .....................................................................................
__all__ = ['clean_data', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Tests the randomize Grady method module."""
import numpy as np
import pytest

from lmpy import Matrix
from lmpy.randomize.grady import (
    all_ones_heuristic,
    all_zeros_heuristic,
    fill_shuffle_reshape_heuristic,
    grady_randomize,
    max_col_or_row_heuristic,
    min_col_or_row_heuristic,
    total_fill_percentage_heuristic,
)
from copy import deepcopy


# .............................................................................
@pytest.fixture(
    scope='session',
    params=[
        None,
        all_ones_heuristic,
        all_zeros_heuristic,
        fill_shuffle_reshape_heuristic,
        max_col_or_row_heuristic,
        min_col_or_row_heuristic,
        total_fill_percentage_heuristic,
    ],
)
def heuristic_func(request):
    """Pytest fixture providing heuristic function.

    Args:
        request (Pytest.Request): A request for a test fixture.

    Yields:
        Method: A heuristic method to use for testing.
    """
    yield request.param


# .............................................................................
class Test_grady_randomize:
    """Test the Grady randomize method."""

    # .....................................
    def _test_pam_randomization(self, pam, heuristic=None):
        """Helper function to test randomization.

        Args:
            pam (Matrix): A presence absence matrix to randomize.
            heuristic (Method): A heuristic method for creating an approximate
                randomization.

        Raises:
            Exception: Raised if the PAM was not randomized.
        """
        if heuristic is not None:
            rand = grady_randomize(pam, approximation_heuristic=heuristic)
        else:
            rand = grady_randomize(deepcopy(pam))
        # Randomize until not the same as original
        count = 10
        while np.all(rand == pam) and count > 0:
            if heuristic is not None:
                rand = grady_randomize(pam, approximation_heuristic=heuristic)
            else:
                rand = grady_randomize(deepcopy(pam))
            count -= 1
        if count <= 0:
            raise Exception('Randomized PAM is same as original')

        # Test that marginal totals are the same
        assert np.all(np.sum(pam, axis=1) == np.sum(rand, axis=1))
        assert np.all(np.sum(pam, axis=0) == np.sum(rand, axis=0))

    # .....................................
    def test_empty_column(self, heuristic_func):
        """Test with a known matrix that has an empty column.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 0, 0, 0], [0, 0, 0, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_empty_row(self, heuristic_func):
        """Test with a known matrix that has an empty row.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 0, 0], [0, 0, 0, 0, 0], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_empty_row_and_column(self, heuristic_func):
        """Test with a known matrix that has an empty row and column.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(
            np.array(
                [
                    [
                        1,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        0,
                    ],
                    [
                        1,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        1,
                        1,
                    ],
                    [
                        1,
                        1,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        0,
                        1,
                        1,
                    ],
                    [
                        1,
                        1,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        1,
                    ],
                    [
                        1,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        1,
                        1,
                    ],
                    [
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        1,
                    ],
                    [
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        1,
                        1,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        0,
                        1,
                        1,
                    ],
                    [
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        1,
                        1,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        1,
                        0,
                        0,
                        1,
                    ],
                    [
                        1,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                    ],
                    [
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        0,
                        1,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        1,
                        0,
                        0,
                        0,
                    ],
                    [
                        1,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                    ],
                    [
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        0,
                    ],
                ]
            )
        )
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_full_column(self, heuristic_func):
        """Test with a known matrix that has a full column.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(
            np.array(
                [[0, 1, 1, 1, 0], [1, 1, 1, 0, 0], [0, 0, 1, 0, 1], [1, 1, 1, 1, 1]]
            )
        )
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_full_row(self, heuristic_func):
        """Test with a known matrix that has a full row.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 1, 1], [0, 0, 1, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_known_matrix(self, heuristic_func):
        """Test with a known matrix.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(
            np.array(
                [[0, 1, 0, 1, 0], [1, 1, 1, 0, 0], [0, 0, 1, 0, 1], [1, 1, 0, 1, 1]]
            )
        )
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_random_large_matrix(self, heuristic_func):
        """Test with a randomly generated large matrix.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(np.random.randint(0, 2, size=(1000, 10000)))
        self._test_pam_randomization(pam, heuristic=heuristic_func)

    # .....................................
    def test_random_matrix(self, heuristic_func):
        """Test with a randomly generated matrix.

        Args:
            heuristic_func (Method): A heuristic method for creating an approximate
                randomization.
        """
        pam = Matrix(np.random.randint(0, 2, size=(10, 10)))
        self._test_pam_randomization(pam, heuristic=heuristic_func)



"""Tool for wrangling a Matrix."""
import argparse

import json

from lmpy import Matrix
from lmpy.data_wrangling.factory import WranglerFactory


DESCRIPTION = '''\
Perform various wrangle operations on a matrix such as subsetting and reordering.'''


# .....................................................................................
def wrangle_matrix(mtx, wranglers):
    """Wrangle the Matrix as configured.

    Args:
        mtx (Matrix): A Matrix to wrangle.
        wranglers (list): A list of tree data wranglers.

    Returns:
        Matrix: The modified Matrix object.
        list: A list of report dictionaries.
    """
    report = []
    for wrangler in wranglers:
        mtx = wrangler.wrangle_matrix(mtx)
        report.append(wrangler.get_report())
    return mtx, report


# .....................................................................................
def cli():
    """Provice a command-line interface to the wrangle matrix tool."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '-r',
        '--report_filename',
        type=str,
        help='File location to write the wrangler report.'
    )
    parser.add_argument(
        'in_matrix_filename', type=str, help='Path to the input Matrix.'
    )
    parser.add_argument(
        'wrangler_configuration_file',
        type=str,
        help='Path to Matrix wrangler configuration.',
    )
    parser.add_argument(
        'out_matrix_filename', type=str, help='Path to the outut Matrix.'
    )
    args = parser.parse_args()

    in_mtx = Matrix.load(args.in_matrix_filename)
    with open(args.wrangler_configuration_file, mode='rt') as in_json:
        wrangler_factory = WranglerFactory()
        wranglers = wrangler_factory.get_wranglers(json.load(in_json))

    wrangled_mtx, report = wrangle_matrix(in_mtx, wranglers)
    wrangled_mtx.write(args.out_matrix_filename)

    if args.report_filename is not None:
        with open(args.report_filename, mode='wt') as report_out:
            json.dump(report, report_out)


# .....................................................................................
__all__ = ['cli', 'wrangle_matrix']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module for the Lifemapper TreeWrapper class.

Todo:
    * Should we provide a method to collapse clades that only have one child?
    * Add method to remove annotations.
    * Move label method out of internal functions.
"""
import os

import dendropy
import numpy as np

from .matrix import Matrix


# .............................
class PhyloTreeKeys:
    """Keys for phylogenetic trees.

    Attributes:
        MTX_IDX (str): The tree attribute indicating the matrix index position
            for a node.
        SQUID (str): The tree attribute indicating a hashed identifier for the
            taxon.
    """

    MTX_IDX = 'mx'  # The matrix index for this node
    SQUID = 'squid'  # This is the LM SQUID (species identifier) for the tip


# .............................................................................
class LmTreeException(Exception):
    """Wrapper around the base Exception class for tree related errors."""

    pass


# .............................................................................
class TreeWrapper(dendropy.Tree):
    """Dendropy Tree wrapper.

    Dendropy tree wrapper that adds a little functionality and improves
    performance of some functions.
    """

    # ..............................
    @classmethod
    def from_base_tree(cls, tree):
        """Creates a TreeWrapper object from a base dendropy.Tree.

        Args:
            tree (:obj:`Tree`): A base dendropy tree object to wrap into a
                TreeWrapper.

        Returns:
            TreeWrapper: The newly wrapped tree.
        """
        return cls.get(data=tree.as_string('nexus'), schema='nexus')

    # ..............................
    @classmethod
    def from_filename(cls, filename):
        """Creates a TreeWrapper object by loading a file.

        Args:
            filename (:obj:`str`): A file path to a tree file that should be loaded.

        Returns:
            TreeWrapper: The newly loaded tree.

        Raises:
            IOError: Raised if the tree file cannot be loaded based on the file
                extension.
        """
        _, tree_ext = os.path.splitext(filename)
        if tree_ext == '.nex':
            tree_schema = 'nexus'
        elif tree_ext == '.xml':
            tree_schema = 'nexml'
        elif tree_ext == '.tre':
            tree_schema = 'newick'
        else:
            raise IOError('Cannot handle tree with extension: {}'.format(tree_ext))
        return cls.get(path=filename, schema=tree_schema)

    # ..............................
    def add_node_labels(self, prefix=None, overwrite=False):
        """Add labels to the nodes in the tree.

        Add labels to the unlabeled nodes in the tree.

        Args:
            prefix (:obj:`str`, optional): If provided, prefix the node labels
                with this string.
            overwrite (:obj:`bool`, optional): Indicates whether existing node
                labels should be overwritten or if they should be maintained.
                Defaults to False.

        Note:
            * This labels nodes the way that R does.
        """
        self._label_tree_nodes(
            self.seed_node, len(self.get_labels()), prefix=prefix, overwrite=overwrite
        )

    # ..............................
    def annotate_tree(
        self,
        annotation_dict,
        annotation_attribute=None,
        label_attribute=None,
        update=False,
    ):
        """Annotates tree tips and nodes.

        Args:
            annotation_dict (:obj:`dict`): A dictionary where the keys
                correspond with the node labels and the value is either, a
                single value, or a dictionary of annotation name keys and
                annotation value values.
            annotation_attribute (:obj:`str` or None, optional): Only used if
                annotation_dict contains single values, this will be the name
                of the annotation added for each node.  Using None or setting
                value to 'label' will change the label of the node.  Defaults
                to None.
            label_attribute (:obj:`str`, optional): Use the value of this
                annotation as the label for the node.  Setting the value to
                'label' or leaving as None will use the label of the node.
                Defaults to None.
            update (:obj:`bool`, optional): If True, update any existing
                annotations with the annotations provided.  Defaults to False.
        """
        label_method = self._get_label_method(label_attribute)

        for node in self.nodes():
            # Get the label of the node
            label = label_method(node)
            if label in annotation_dict.keys():
                node_annotation = annotation_dict[label]

                # Assume annotations are in a dictionary
                try:
                    for (ann_name, ann_value) in node_annotation.items():
                        self._annotate_node(node, ann_name, ann_value, update=update)
                except Exception:
                    # Annotation is a single value
                    self._annotate_node(
                        node, annotation_attribute, node_annotation, update=update
                    )

    # ..............................
    def annotate_tree_tips(
        self, attribute_name, annotation_pairs, label_attribute='label', update=False
    ):
        """Annotates the tips of the tree.

        Deprecated:
            Update to use annotate_tree.

        Args:
            attribute_name (:obj:`str`): The name of the annotation attribute to add.
            annotation_pairs (:obj:`dict`): A dictionary of label keys with annotation
                values.
            label_attribute (:obj:`str`, optional): If this is provided, use this
                annotation attribute as the key instead of the label. Defaults to
                'label'.
            update (:obj:`bool`, optional): Defaults to False.  Indicates if existing
                annotations should be updated.
        """
        label_method = self._get_label_method(label_attribute)

        for taxon in self.taxon_namespace:
            try:
                # label = getattr(taxon, labelAttribute)
                label = label_method(taxon)

                if taxon.annotations.get_value(attribute_name) is not None:
                    if update:
                        # Remove existing values
                        for ann in taxon.annotations.findall(name=attribute_name):
                            taxon.annotations.remove(ann)
                        # Set new value
                        taxon.annotations.add_new(
                            attribute_name, annotation_pairs[label]
                        )
                else:
                    taxon.annotations.add_new(attribute_name, annotation_pairs[label])
            except KeyError:
                # Pass if label is not found in the dictionary, otherwise fail
                pass

    # ..............................
    def get_annotations(self, annotation_attribute):
        """Gets a list of (label, annotation) pairs.

        Args:
            annotation_attribute (:obj:`str`): The annotation attribute to retrieve.

        Returns:
            list: A list of annotations.
        """
        annotations = []
        for taxon in self.taxon_namespace:
            att = taxon.annotations.get_value(annotation_attribute)
            annotations.append((taxon.label, att))
        return annotations

    # ..............................
    def get_distance_matrix(self, label_attribute='label', ordered_labels=None):
        """Gets a Matrix object of phylogenetic distances.

        Get a Matrix object of phylogenetic distances between tips using a
        lower memory footprint.

        Args:
            label_attribute (:obj:`str`, optional): The attribute of the tips to use as
                labels for the matrix.  Defaults to 'label'.
            ordered_labels (:obj:`list` of :obj:`str`, optional): If provided, use this
                order of labels.

        Returns:
            Matrix: A distance matrix from each tip to each of the other tips
                in the tree.
        """
        label_method = self._get_label_method(label_attribute)

        # Get list of labels
        if ordered_labels is None:
            ordered_labels = []
            for taxon in self.taxon_namespace:
                ordered_labels.append(label_method(taxon))

        label_lookup = {ordered_labels[i]: i for i in range(len(ordered_labels))}

        dist_mtx = np.zeros((len(ordered_labels), len(ordered_labels)), dtype=float)

        # Build path lookup dictionary
        path_lookups = {}
        for taxon in self.taxon_namespace:
            test_node = self.find_node_for_taxon(taxon)
            edge_lengths = []
            while test_node is not None:
                if test_node.edge_length is not None:
                    # If this is still too big, drop id and do some logic with
                    #    lengths
                    edge_lengths.append((id(test_node), test_node.edge_length))
                test_node = test_node.parent_node
            path_lookups[taxon.label] = edge_lengths

        num_taxa = len(self.taxon_namespace)
        for i_1 in range(num_taxa - 1):
            taxon1 = self.taxon_namespace[i_1]

            label = label_method(taxon1)
            # Check for matrix index
            idx1 = label_lookup[label]

            # Build path to root for taxon 1
            # path_labels = []
            o_dist = 0.0
            t_path = path_lookups[taxon1.label]
            t_labels = []
            for label, p_dist in t_path:
                o_dist += p_dist
                t_labels.append(label)

            for i_2 in range(i_1, num_taxa):
                taxon2 = self.taxon_namespace[i_2]

                idx2 = label_lookup[label_method(taxon2)]

                # Initialize distance for these two taxa
                dist = o_dist

                # Loop through path back to root
                t2_path = path_lookups[taxon2.label]

                for label, p_dist in t2_path:
                    if label in t_labels:
                        dist -= p_dist
                    else:
                        dist += p_dist

                # mrca = pdm.mrca(taxon1, taxon2)
                # dist = pdm.patristic_distance(taxon1, taxon2)
                dist_mtx[idx1, idx2] = dist
                dist_mtx[idx2, idx1] = dist

        distance_matrix = Matrix(
            dist_mtx, headers={'0': ordered_labels, '1': ordered_labels}
        )
        return distance_matrix

    # ..............................
    def get_distance_matrix_dendropy(
        self, label_attribute='label', ordered_labels=None
    ):
        """Gets a Matrix object of phylogenetic distances between tips.

        Gets the distance matrix between each tip using Dendropy.

        Args:
            label_attribute (:obj:`str`, optional): The attribute of the tips to use as
                labels for the matrix.  Defaults to 'label'.
            ordered_labels (:obj:`list` of :obj:`str`, optional): If provided, use this
                order of labels.

        Note:
            This method may require a significant amount of memory for large trees.
                The `get_distance_matrix` method has a smaller memory footprint and
                works at nearly the same speed.

        Returns:
            Matrix: A distance matrix from each tip to each of the other tips in the
                tree.
        """
        label_method = self._get_label_method(label_attribute)

        # Get list of labels
        if ordered_labels is None:
            ordered_labels = []
            for taxon in self.taxon_namespace:
                ordered_labels.append(label_method(taxon))

        label_lookup = {ordered_labels[i]: i for i in range(len(ordered_labels))}

        dist_mtx = np.zeros((len(ordered_labels), len(ordered_labels)), dtype=float)

        pdm = self.phylogenetic_distance_matrix()

        for taxon1 in self.taxon_namespace:
            label = label_method(taxon1)
            # Check for matrix index
            idx1 = label_lookup[label]

            for taxon2 in self.taxon_namespace:
                idx2 = label_lookup[label_method(taxon2)]
                # mrca = pdm.mrca(taxon1, taxon2)
                dist = pdm.patristic_distance(taxon1, taxon2)
                dist_mtx[idx1, idx2] = dist

        distance_matrix = Matrix(
            dist_mtx, headers={'0': ordered_labels, '1': ordered_labels}
        )
        return distance_matrix

    # ..............................
    def get_labels(self):
        """Gets tip labels for a clade.

        Note:
            Bottom-up order.

        Returns:
            A list of taxon labels for the taxa in the tree.
        """
        labels = []
        for taxon in self.taxon_namespace:
            labels.append(taxon.label)

        labels.reverse()
        return labels

    # ..............................
    def get_variance_covariance_matrix(
        self, label_attribute='label', ordered_labels=None
    ):
        """Gets a Matrix object of variance / co-variance for tips in tree.

        Args:
            label_attribute (:obj:`str`, optional): The attribute of the tips to use as
                labels for the matrix.  Defaults to 'label'.
            ordered_labels (:obj:`list` of :obj:`str`, optional): If provided, use this
                order of labels.

        Returns:
            Matrix: A matrix of variance / co-variance values for the tips in
                the tree.

        Raises:
            LmTreeException: If the tree does not have branch lengths.
        """
        if not self.has_branch_lengths():
            raise LmTreeException('Cannot create VCV without branch lengths')

        label_method = self._get_label_method(label_attribute)

        # Get list of labels
        if ordered_labels is None:
            ordered_labels = []
            for taxon in self.taxon_namespace:
                ordered_labels.append(label_method(taxon))

        label_lookup = {ordered_labels[i]: i for i in range(len(ordered_labels))}

        n = len(ordered_labels)
        vcv = np.zeros((n, n), dtype=float)

        edges = []
        for edge in self.postorder_edge_iter():
            edges.append(edge)

        edges.reverse()

        for edge in edges:
            # TODO: Evaluate if el assignment can throw error
            el = edge.head_node.distance_from_root()
            if el is not None:
                child_nodes = edge.head_node.child_nodes()
                if len(child_nodes) == 0:
                    idx = label_lookup[label_method(edge.head_node.taxon)]
                    vcv[idx, idx] = edge.head_node.distance_from_root()
                else:
                    left_child, right_child = edge.head_node.child_nodes()
                    left_tips = [
                        label_lookup[label_method(tip_node.taxon)]
                        for tip_node in left_child.leaf_nodes()
                    ]
                    right_tips = [
                        label_lookup[label_method(tip_node.taxon)]
                        for tip_node in right_child.leaf_nodes()
                    ]
                    # if len(leftTips) > 1 and len(rightTips) > 1:
                    for left in left_tips:
                        for right in right_tips:
                            vcv[left, right] = vcv[right, left] = el

            for node in self.leaf_nodes():
                idx = label_lookup[label_method(node.taxon)]
                vcv[idx, idx] = node.distance_from_root()

        vcv_matrix = Matrix(vcv, headers={'0': ordered_labels, '1': ordered_labels})
        return vcv_matrix

    # ..............................
    def has_branch_lengths(self):
        """Returns a boolean indicating if the entire tree has branch lengths.

        Returns:
            bool: An indication if the tree has branch lengths.
        """
        try:
            self.minmax_leaf_distance_from_root()
            return True
        except Exception:
            return False

    # ..............................
    def has_polytomies(self):
        """Returns boolean indicating if the tree has polytomies.

        Returns:
            bool: An indication if the tree has any polytomies.
        """
        for n in self.nodes():
            if len(n.child_nodes()) > 2:
                return True
        return False

    # ..............................
    def is_binary(self):
        """Checks if the tree is binary.

        Returns:
            bool: An indication if the tree is binary.

        Note:
            * Checks that every clade has either zero or two children.
        """
        for n in self.nodes():
            if not len(n.child_nodes()) in [0, 2]:
                return False
        return True

    # ..............................
    def is_ultrametric(self, rel_tol=1e-03):
        """Checks if the tree is ultrametric.

        Args:
            rel_tol (:obj:`float`): The relative tolerance to determine if the min and
                max are equal.  We will say they are equal if they are 99.9%.

        Returns:
            bool: Returns true if the distance from the root to each tip is the
                same (within the tolerance interval).

        Note:
            * To be ultrametric, the branch length from root to tip must be
                equal for all tips.
        """
        try:
            min_bl, max_bl = self.minmax_leaf_distance_from_root()
            return bool(np.isclose(min_bl, max_bl, rtol=rel_tol))
        except TypeError:
            pass
        return False

    # ..............................
    def prune_tips_without_attribute(self, search_attribute=PhyloTreeKeys.MTX_IDX):
        """Prunes the tree of any tips that don't have the specified attribute.

        Args:
            search_attribute (:obj:`str`, optional): The attribute to look for when
                pruning tips in the tree.  Defaults to PhyloTreeKeys.MTX_IDX.
        """
        prune_taxa = []
        for taxon in self.taxon_namespace:
            val = taxon.annotations.get_value(search_attribute)
            if val is None:
                prune_taxa.append(taxon)
        self.prune_taxa(prune_taxa)
        self.purge_taxon_namespace()

    # ..............................
    def _annotation_method(self, label_attribute):
        """Use the label attribute as the node label.

        Args:
            label_attribute (:obj:`str`): The annotation to use as the label for the
                nodes in the tree.

        Returns:
            Method: A method for retrieving the label for a taxon.
        """

        def label_method(taxon):
            """Get the label of a taxon.

            Args:
                taxon (:obj:`Taxon`): A taxon object to get the label for.

            Returns:
                str: The taxon's label.
            """
            return taxon.annotations.get_value(label_attribute)

        return label_method

    # ..............................
    def _annotate_node(
        self, node, annotation_attribute, annotation_value, update=False
    ):
        """Annotates a node with the given value.

        Args:
            node (:obj:`Node`): A node to add an annotation to.
            annotation_attribute (:obj:`str`): The annotation attribute to add.  If
                None or 'label', update the node label.
            annotation_value (:obj:`object`): The value of the annotation.
            update (:obj:`bool`, optional): If True, update existing attribute.
                Defaults to False.
        """
        if annotation_attribute is None or annotation_attribute.lower() == 'label':
            try:
                node.taxon.label = annotation_value
            except Exception:
                node.label = annotation_value
        else:
            if node.annotations.get_value(annotation_attribute) is not None:
                if update:
                    # Remove existing annotations
                    for ann in node.annotations.findall(name=annotation_attribute):
                        node.annotations.remove(ann)
                    node.annotations.add_new(annotation_attribute, annotation_value)
            else:
                node.annotations.add_new(annotation_attribute, annotation_value)

    # ..............................
    def _get_label_method(self, label_attribute):
        """Gets the function to be used for retrieving labels.

        Args:
            label_attribute (:obj:`str`): An annotation name, 'label', or None used to
                determine which method to use to retrieve the label of a node.

        Returns:
            Function for labeling nodes.
        """
        if label_attribute is None or label_attribute.lower() == 'label':
            return self._label_method
        return self._annotation_method(label_attribute)

    # ..............................
    def _label_tree_nodes(self, node, i, prefix=None, overwrite=False):
        """Private function to do the work when labeling nodes.

        Args:
            node (:obj:`Node`): A node to label.
            i (:obj:`int`): A count of the number of previously labeled nodes.
            prefix (:obj:`str`, optional): A prefix to use when labeling nodes
                resulting in labels like 'prefix_0'.  Defaults to None and no prefix.
            overwrite (:obj:`bool`, optional): Should node labels be overwritten.
                Defaults to False.

        Returns:
            int: The number of nodes already labeled in the tree.

        Note:
            * Recursive.
        """
        cn = node.child_nodes()

        # If we have children, label and recurse
        if len(cn) > 0:
            if node.label is None or overwrite:
                if prefix is not None:
                    node.label = '{}{}'.format(prefix, i)
                else:
                    node.label = str(i)
                i += 1
            # Loop through children and label nodes
            for child in cn:
                i = self._label_tree_nodes(child, i, prefix=prefix, overwrite=overwrite)
        # Return the current i value
        return i

    # ..............................
    @staticmethod
    def _label_method(node):
        """Use the label of the node or taxon for the label.

        Args:
            node (:obj:`Node`): The node to get the label for.

        Returns:
            str: If the node or the node's taxon has a label, return it.
            None: If the node and it's taxon do not have labels.
        """
        # If the node (or taxon) has a label, return it
        if node.label is not None:
            return node.label
        # Try to return the label of the taxon object if it has one
        try:
            return node.taxon.label
        except Exception:
            # Fall back to returning None
            return None


# .............................................................................
# Set the module for the classes to be lmpy
PhyloTreeKeys.__module__ = 'lmpy'
TreeWrapper.__module__ = 'lmpy'

__all__ = ['LmTreeException', 'PhyloTreeKeys', 'TreeWrapper']



"""Module containing attribute modifier occurrence data wrangler."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class AttributeModifierWrangler(_OccurrenceDataWrangler):
    """Modifies a point attribute according to a function."""
    name = 'AttributeModifierWrangler'
    version = '1.0'

    # .......................
    def __init__(self, attribute_name, attribute_func, **params):
        """Constructor for AttributeModifierWrangler class.

        Args:
            attribute_name (str): The name of the attribute to modify.
            attribute_func (Method): A function to generate values for a point.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.attribute_name = attribute_name
        self.attribute_func = attribute_func
        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _modify_point(self, point):
        """Update point attributes.

        Args:
            point (Point): A point to modify.

        Returns:
            Point, bool: Modified point and boolean if point was modified.
        """
        point.set_attribute(self.attribute_name, self.attribute_func(point))

        return point, True



"""Module containing occurrence data wranglers for filtering points."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
class DecimalPrecisionFilter(_OccurrenceDataWrangler):
    """Get an occurrence data wrangler for filtering by decimal precision."""
    name = 'DecimalPrecisionFilter'
    version = '1.0'

    # .......................
    def __init__(self, decimal_places, **params):
        """Get an occurrence data wrangler for filtering by decimal precision.

        Args:
            decimal_places (int): Only keep points with at least this many decimal
                places of precision.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.decimal_places = decimal_places
        _OccurrenceDataWrangler.__init__(self, **params)

    # .......................
    def _pass_condition(self, point):
        """Assessment condition for a point.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passed assessment.
        """
        lat_str = str(point.y)
        lon_str = str(point.x)

        try:
            lat_decimals = len(lat_str) - lat_str.index('.') - 1
            lon_decimals = len(lon_str) - lon_str.index('.') - 1
        except ValueError:
            # TODO: Handle numbers with 'e' example: 1e-05
            return False
        return min([lat_decimals, lon_decimals]) >= self.decimal_places



"""This module tests the lmpy.statistics.mcpa.py module."""
import numpy as np

from lmpy import Matrix, TreeWrapper
from lmpy.data_preparation.tree_encoder import TreeEncoder
from lmpy.statistics.mcpa import get_p_values, mcpa


ROUND_POSITION = 7


# .............................................................................
def _make_ultrametric_helper(species, max_branch_length):
    """Helper function for creating an ultrametric tree.

    Args:
        species (list): List of species to include in the tree.
        max_branch_length (numeric): Maximum branch length.

    Returns:
        tuple: Tuple of tree root node and height of the node.
    """
    # If length of list == 1, bounce
    if species and len(species) == 1:
        node = species[0].replace(' ', '_')
        node_height = 0.0
    else:
        # Split and recurse
        if len(species) > 2:
            split_pos = np.random.randint(1, len(species) - 1)
        else:
            split_pos = 1
        node_a, node_height_a = _make_ultrametric_helper(
            species[:split_pos], max_branch_length
        )
        node_b, node_height_b = _make_ultrametric_helper(
            species[split_pos:], max_branch_length
        )
        # Get the maximum node height and add this branch length
        node_height = round(
            max(node_height_a, node_height_b)
            + (max_branch_length * np.random.random()),
            ROUND_POSITION,
        )
        # Make sure node heights are equal for each branch by subtracting
        #    from node height
        node = '({}:{},{}:{})'.format(
            node_a, node_height - node_height_a, node_b, node_height - node_height_b
        )
    # Return branch and height
    return (node, node_height)


# .............................................................................
def _get_random_pam_and_tree(
    num_species, num_sites, fill_percentage, max_branch_length, num_mismatches=0
):
    """Get a random PAM and matching tree.

    Args:
        num_species (int): The number of species in the PAM / Tree.
        num_sites (int): The number of sites in the PAM.
        fill_percentage (float): The percentage of cells that are present in the PAM.
        max_branch_length (float): The maximum branch length for each branch in the
            tree.
        num_mismatches (int): A number of mismatches to include between PAM and tree.

    Returns:
        tuple: A tuple of the presence-absence Matrix and the associated TreeWrapper.
    """
    site_headers = ['Site {}'.format(site_idx) for site_idx in range(num_sites)]
    pam_species = []
    tree_species = []
    for sp_idx in range(num_species):
        sp = 'Species{}'.format(sp_idx)
        pam_species.append(sp)
        tree_species.append(sp)

    # Add a few species to PAM species and tree species
    for sp_idx in range(num_mismatches):
        pam_species.append('PamSpecies {}'.format(sp_idx))
        tree_species.append('TreeSpecies {}'.format(sp_idx))

    np.random.shuffle(pam_species)
    np.random.shuffle(tree_species)

    pam = Matrix(
        (np.random.random((num_sites, num_species)) < fill_percentage).astype(int),
        headers={'0': site_headers, '1': pam_species},
    )

    if max_branch_length:
        tree_data = '{};'.format(
            _make_ultrametric_helper(tree_species, max_branch_length)[0]
        )
    else:
        tmp = []
        tmp.extend(tree_species)
        while len(tmp) > 1:
            sp_1 = tmp.pop()
            sp_2 = tmp.pop()
            tmp.append('({},{})'.format(sp_1, sp_2))
        tree_data = '{};'.format(tmp[0])
    tree = TreeWrapper.get(data=tree_data, schema='newick')
    tree.annotate_tree_tips('squid', {sp: sp for sp in tree_species})
    tree.annotate_tree_tips('mx', {sp: idx for idx, sp in enumerate(pam_species)})
    return (pam, tree)


# .............................................................................
def _create_env_and_biogeo_matrices(pam, num_env_cols, num_bg_cols):
    """Create environment and bio geo matrices matching pam sites.

    Args:
        pam (Matrix): A presence-absence Matrix.
        num_env_cols (int): The number of environment layers to generate.
        num_bg_cols (int): The number of biogeographic hypotheses to generate.

    Returns:
        tuple: A tuple of enviroment Matrix and biogeo hypotheses Matrix.
    """
    bg_mtx = Matrix(
        np.random.randint(-1, 2, (pam.shape[0], num_bg_cols)),
        headers={
            '0': pam.get_row_headers(),
            '1': ['BG - {}'.format(i) for i in range(num_bg_cols)],
        },
    )
    env_mtx = Matrix(
        np.random.uniform(0.0, 100.0, (pam.shape[0], num_env_cols)),
        headers={
            '0': pam.get_row_headers(),
            '1': ['Env - {}'.format(i) for i in range(num_env_cols)],
        },
    )
    return (env_mtx, bg_mtx)


# .............................................................................
class Test_mcpa:
    """Test various individual metrics."""

    # ............................
    def test_mcpa_with_branch_lengths(self):
        """Test MCPA with branch lengths."""
        # Need PAM, Encoded tree, Env matrix, BG matrix
        pam, tree = _get_random_pam_and_tree(10, 20, 0.3, 1.0)
        phylo_mtx = TreeEncoder(tree, pam).encode_phylogeny()
        env_mtx, bg_mtx = _create_env_and_biogeo_matrices(pam, 10, 3)
        mcpa(pam, phylo_mtx, env_mtx, bg_mtx)

    # ............................
    def test_mcpa_no_branch_lengths(self):
        """Test MCPA no branch lengths."""
        # Need PAM, Encoded tree, Env matrix, BG matrix
        pam, tree = _get_random_pam_and_tree(10, 20, 0.3, None)
        phylo_mtx = TreeEncoder(tree, pam).encode_phylogeny()
        env_mtx, bg_mtx = _create_env_and_biogeo_matrices(pam, 10, 3)
        mcpa(pam, phylo_mtx, env_mtx, bg_mtx)


# .............................................................................
class Test_get_p_values:
    """Test the get_p_values method.

    Note:
        Deprecate this in favor of using running stats version.
    """

    # ............................
    def test_single_permutation(self):
        """Test with a single test permutation."""
        obs_mtx = Matrix(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        test_mtx = Matrix(
            np.array([[9, 8, 7], [6, 5, 4], [3, 2, 1]]).reshape((3, 3, 1))
        )
        p_vals = get_p_values(obs_mtx, [test_mtx])[:, :, 0]
        p_test_mtx = Matrix(np.array([[1, 1, 1], [1, 1, 0], [0, 0, 0]]))
        assert np.all(p_vals == p_test_mtx)

    # ............................
    def test_single_permutation_3dim(self):
        """Test with a single test permutation."""
        obs_mtx = Matrix(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        test_mtx = Matrix(np.array([[9, 8, 7], [6, 5, 4], [3, 2, 1]]))
        p_vals = get_p_values(obs_mtx, [test_mtx])[:, :, 0]
        p_test_mtx = Matrix(np.array([[1, 1, 1], [1, 1, 0], [0, 0, 0]]))
        assert np.all(p_vals == p_test_mtx)

    # ............................
    def test_multiple_permutations(self):
        """Test with multiple test permutations."""
        obs_mtx = Matrix(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        test_mtxs = [
            Matrix(np.array([[9, 8, 7], [6, 5, 4], [3, 2, 1]])),
            Matrix(np.array([[5, 5, 5], [5, 5, 5], [5, 5, 5]])),
            Matrix(np.array([[3, 3, 3], [6, 6, 6], [8, 8, 8]])),
            Matrix(np.array([[3, 2, 1], [6, 5, 4], [9, 8, 7]])),
        ]
        p_vals = get_p_values(obs_mtx, test_mtxs, num_permutations=len(test_mtxs))[
            :, :, 0
        ]
        p_test_mtx = Matrix(np.array([[1, 1, 0.75], [1, 1, 0.25], [0.5, 0.5, 0]]))
        assert np.all(p_vals == p_test_mtx)



"""Module containing Tree Data Wrangler base class."""
from lmpy.data_wrangling.base import _DataWrangler


# .....................................................................................
class _TreeDataWrangler(_DataWrangler):
    """Tree data wrangler base class."""
    name = '_TreeDataWrangler'

    # .......................
    def __init__(self, **params):
        """Constructor for _TreeDataWrangler base class.

        Args:
            **params (dict): Named parameters to pass to _DataWrangler base class.
        """
        _DataWrangler.__init__(self, **params)
        self._modified = 0
        self._purged = 0

    # .......................
    def _report_tip(self, modified=False, purged=False):
        """Report what is done on a tree tip.

        Args:
            modified (bool): Was the tip modified.
            purged (bool): Was the tip purged.
        """
        if modified:
            self._modified += 1

        if purged:
            self._purged += 1

    # .......................
    def get_report(self):
        """Get the report of the wrangler's activities.

        Returns:
            dict: Wrangler report information.
        """
        report = _DataWrangler.get_report(self)
        report['purged'] = self._purged
        report['modified'] = self._modified
        return report

    # .......................
    def wrangle_tree(self, tree):
        """Wrangle a tree.

        Args:
            tree (TreeWrapper): A phylogenetic tree to wrangle.

        Raises:
            NotImplementedError: This method is not implemeneted for the base class.
        """
        raise NotImplementedError('wrangle_matrix not implemented on base class.')



"""Module containing occurrence data wranglers for modifying point data."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler
from lmpy.data_wrangling.common.accepted_name_wrangler import (
    _AcceptedNameWrangler,
    resolve_names_gbif,
)


# .....................................................................................
class AcceptedNameOccurrenceWrangler(_OccurrenceDataWrangler, _AcceptedNameWrangler):
    """Modifies the species_name to the "accepted" taxon name for the species."""
    name = 'AcceptedNameOccurrenceWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        name_map=None,
        name_resolver=None,
        store_original_attribute=None,
        **params
    ):
        """Constructor for AcceptedNameModifier class.

        Args:
            name_map (dict): A map of original name to accepted name.
            name_resolver (str or Method): If provided, use this method for getting new
                accepted names.  If set to 'gbif', use GBIF name resolution.
            store_original_attribute (str or None): A new attribute to store the
                original taxon name.
            **params (dict): Keyword parameters to pass to _OccurrenceDataWrangler.
        """
        _OccurrenceDataWrangler.__init__(self, **params)

        if isinstance(name_resolver, str) and name_resolver.lower() == 'gbif':
            name_resolver = resolve_names_gbif
        _AcceptedNameWrangler.__init__(
            self,
            name_map=name_map,
            name_resolver=name_resolver
        )

        self.store_original_attribute = store_original_attribute

    # .......................
    def _pass_condition(self, point):
        """Determine if a point has an accepted name.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: An indication if the point passed the test.
        """
        if point.species_name is None or len(point.species_name) == 0:
            return False
        return True

    # .......................
    def _modify_point(self, point):
        """Update taxon name if necessary.

        Args:
            point (Point): A point to modify.

        Returns:
            Point, bool: Modified (if needed) point and boolean if point was modified.
        """
        is_modified = False
        acc_name = self.resolve_names([point.species_name])[point.species_name]

        # If we should keep original, add a new attribute
        if self.store_original_attribute is not None:
            point.set_attribute(self.store_original_attribute, point.species_name)
            is_modified = True

        if point.species_name != acc_name:
            point.species_name = acc_name
            is_modified = True

        return point, is_modified



"""Tool for creating PAM statistics."""
import argparse
import json

from lmpy.matrix import Matrix
from lmpy.tree import TreeWrapper
from lmpy.statistics.pam_stats import PamStats

DESCRIPTION = 'Compute statistics for a PAM and optionally a tree.'


# .....................................................................................
def cli():
    """Provide a command-line tool for computing statistics."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--tree_filename', type=str, help='Path to matching tree.')
    parser.add_argument(
        '--tree_matrix', type=str, nargs=3, help='Path to tree matrix.'
    )

    # Output matrices
    parser.add_argument(
        '--covariance_matrix',
        type=str,
        help='Path to write covariance matrix if desired.',
    )
    parser.add_argument(
        '--diversity_matrix',
        type=str,
        help='Path to write diversity matrix if desired.',
    )
    parser.add_argument(
        '--site_stats_matrix',
        type=str,
        help='Path to write site statistics matrix if desired.',
    )
    parser.add_argument(
        '--species_stats_matrix',
        type=str,
        help='Path to write species statistics matrix if desired.',
    )

    # PAM
    parser.add_argument('pam_filename', type=str, help='Path to PAM file.')

    args = parser.parse_args()

    tree = tree_matrix = node_heights_matrix = tip_lengths_matrix = None
    pam = Matrix.load(args.pam_filename)

    if args.tree_filename is not None:
        tree = TreeWrapper.from_filename(args.tree_filename)

    if args.tree_matrix is not None:
        tree_matrix = Matrix.load(args.tree_matrix[0])
        node_heights_matrix = Matrix.load(args.tree_matrix[1])
        tip_lengths_matrix = Matrix.load(args.tree_matrix[2])

    stats = PamStats(
        pam,
        tree=tree,
        tree_matrix=tree_matrix,
        node_heights_matrix=node_heights_matrix,
        tip_lengths_matrix=tip_lengths_matrix,
    )

    # Write requested stats
    if args.covariance_matrix is not None:
        with open(args.covariance_matrix, mode='wt') as out_json:
            json.dump(stats.calculate_covariance_statistics(), out_json)

    if args.diversity_matrix is not None:
        stats.calculate_diversity_statistics().write(args.diversity_matrix)

    if args.site_stats_matrix is not None:
        stats.calculate_site_statistics().write(args.site_stats_matrix)

    if args.species_stats_matrix is not None:
        stats.calculate_species_statistics().write(args.species_stats_matrix)


# .....................................................................................
__all__ = ['cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing a coordinate converter modifier."""
from copy import deepcopy

from osgeo import osr

from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler


# .....................................................................................
def get_srs_for_epsg(epsg):
    """Get a osr.SpatialReference for the specified EPSG code.

    Args:
        epsg (int): An EPSG code to get the SpatialReference for.

    Returns:
        SpatialReference: A spatial reference object for the EPSG.
    """
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)

    try:
        # So we can use GDAL 2 or 3
        # See https://github.com/OSGeo/gdal/issues/1546
        srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    except Exception as err:  # pragma: no cover
        print('Error setting mapping strategy')
        print(err)

    return srs


# .....................................................................................
class CoordinateConverterWrangler(_OccurrenceDataWrangler):
    """Tool for converting fron one coordinate system to another via ESPG."""
    name = 'CoordinateConverterWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        target_epsg,
        source_epsg=None,
        epsg_attribute=None,
        original_x_attribute=None,
        original_y_attribute=None,
        **params
    ):
        """Constructor for CoordinateConverterModifier class.

        Args:
            target_epsg (int): Target map projection specified by EPSG code.
            source_epsg (int or None): Source map projection specified by EPSG code.
            epsg_attribute (str or None): A point attribute containing EPSG code.
            original_x_attribute (str or None): An attribute to store the original x
                value.
            original_y_attribute (str or None): An attribute to store the original y
                value.
            **params (dict): Extra parameters to be sent to the base class.
        """
        self.source_epsg = source_epsg
        self.epsg_attribute = epsg_attribute
        self.original_x_attribute = original_x_attribute
        self.original_y_attribute = original_y_attribute

        _OccurrenceDataWrangler.__init__(self, **params)
        self.target_sr = get_srs_for_epsg(target_epsg)
        self.transforms = {}

    # .......................
    def _modify_point(self, point):
        """Transform the point coordinates.

        Args:
            point (Point): The point object to modify.

        Returns:
            Point and bool: A modified (or not) Point and boolean indicating if it was
                modified.
        """
        if self.source_epsg is not None:
            epsg = int(self.source_epsg)
        else:
            epsg = int(point.get_attribute(self.epsg_attribute))

        if epsg in self.transforms.keys():
            transform = self.transforms[epsg]
        else:
            source_sr = get_srs_for_epsg(epsg)
            transform = osr.CoordinateTransformation(source_sr, self.target_sr)
            self.transforms[epsg] = transform

        new_x, new_y, _ = transform.TransformPoint(point.x, point.y)
        pt = deepcopy(point)
        pt.x = new_x
        pt.y = new_y
        # If we should keep the original values, do so
        if all(
            [
                self.original_x_attribute is not None,
                self.original_y_attribute is not None
            ]
        ):
            pt.set_attribute(self.original_x_attribute, point.x)
            pt.set_attribute(self.original_y_attribute, point.y)

        return pt, True

    # .......................
    def _pass_condition(self, point):
        """Pass condition for coordinate conversion.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passes the test condition.
        """
        if point.x is None or point.y is None:
            return False
        return True



"""Module containing occurrence data wranglers for filtering points."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler
from lmpy.spatial import SpatialIndex


# .....................................................................................
class DisjointGeometriesFilter(_OccurrenceDataWrangler):
    """Get an occurrence data wrangler for filtering by disjoint geometries."""
    name = 'DisjointGeometriesFilter'
    version = '1.0'

    # .......................
    def __init__(self, geometry_wkts, **params):
        """Get an occurrence data wrangler for filtering by disjoint geometries.

        Args:
            geometry_wkts (list of str): A list of geometry WKTs to check against.
            **params (dict): Extra parameters to be sent to the base class.
        """
        _OccurrenceDataWrangler.__init__(self, **params)
        self.geom_index = SpatialIndex()
        i = 0
        for wkt in geometry_wkts:
            self.geom_index.add_feature(i, wkt, {"feature_id": i})
            i += 1

    # .......................
    def _pass_condition(self, point):
        """Assesment of a point to see if it passes the condition.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passes assessment.
        """
        hits = self.geom_index.search(point.x, point.y)
        return not bool(hits)



"""This module tests the lmpy.statistics.pam_stats.py module."""
import numpy as np
import pytest

from lmpy import Matrix, TreeWrapper
import lmpy.statistics.pam_stats as stats


ROUND_POSITION = 7


# .............................................................................
def _make_ultrametric_helper(species, max_branch_length):
    """Helper function for generating an ultrametric tree.

    Args:
        species (list): List of species to include.
        max_branch_length (numeric): The maximum branch length for this branch.

    Returns:
        tuple: A tuple of a tree root node and the height.
    """
    # If length of list == 1, bounce
    if species and len(species) == 1:
        node = species[0].replace(' ', '_')
        node_height = 0.0
    else:
        # Split and recurse
        if len(species) > 2:
            split_pos = np.random.randint(1, len(species) - 1)
        else:
            split_pos = 1
        node_a, node_height_a = _make_ultrametric_helper(
            species[:split_pos], max_branch_length
        )
        node_b, node_height_b = _make_ultrametric_helper(
            species[split_pos:], max_branch_length
        )
        # Get the maximum node height and add this branch length
        node_height = round(
            max(node_height_a, node_height_b)
            + (max_branch_length * np.random.random()),
            ROUND_POSITION,
        )
        # Make sure node heights are equal for each branch by subtracting
        #    from node height
        node = '({}:{},{}:{})'.format(
            node_a, node_height - node_height_a, node_b, node_height - node_height_b
        )
    # Return branch and height
    return (node, node_height)


# .............................................................................
def get_random_pam_and_tree(
    num_species, num_sites, fill_percentage, max_branch_length, num_mismatches=0
):
    """Get a random PAM and matching tree.

    Args:
        num_species (int): The number of species in the PAM / Tree.
        num_sites (int): The number of sites in the PAM.
        fill_percentage (float): The percentage of cells that are present in
            the PAM.
        max_branch_length (float): The maximum branch length for each branch in
            the tree.
        num_mismatches (int): A number of mismatches to include between PAM and
            tree.

    Returns:
        tuple: Return a tuple of the PAM matrix and the tree generated.
    """
    site_headers = ['Site {}'.format(site_idx) for site_idx in range(num_sites)]
    pam_species = []
    tree_species = []
    for sp_idx in range(num_species):
        sp = 'Species {}'.format(sp_idx)
        pam_species.append(sp)
        tree_species.append(sp)

    # Add a few species to PAM species and tree species
    for sp_idx in range(num_mismatches):
        pam_species.append('PamSpecies {}'.format(sp_idx))
        tree_species.append('TreeSpecies {}'.format(sp_idx))

    np.random.shuffle(pam_species)
    np.random.shuffle(tree_species)

    pam = Matrix(
        (np.random.random((num_sites, num_species)) < fill_percentage).astype(int),
        headers={'0': site_headers, '1': pam_species},
    )

    tree_data = '({});'.format(
        _make_ultrametric_helper(tree_species, max_branch_length)[0]
    )
    tree = TreeWrapper.get(data=tree_data, schema='newick')
    tree.annotate_tree_tips('squid', {sp: sp for sp in tree_species})
    return (pam, tree)


# .............................................................................
class Test_Metrics:
    """Test various individual metrics."""

    # ............................
    def test_covariance_metrics(self):
        """Test the covariance metrics."""
        pam, _ = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        sigma_sites = stats.sigma_sites(pam)
        assert sigma_sites.shape == (20, 20)

        sigma_species = stats.sigma_species(pam)
        assert sigma_species.shape == (10, 10)

    # ............................
    def test_diversity_metrics(self):
        """Test the diversity metrics."""
        pam, _ = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        metrics = [
            ('c-score', stats.c_score),
            ('lande', stats.lande),
            ('legendre', stats.legendre),
            ('num sites', stats.num_sites),
            ('num species', stats.num_species),
            ('whittaker', stats.whittaker),
        ]
        for name, func in metrics:
            print(name)
            metric = func(pam)
            assert isinstance(metric, (int, float))

    # ............................
    def test_site_matrix_metrics(self):
        """Test site metrics that take a matrix as input."""
        pam, _ = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        metrics = [
            ('alpha', stats.alpha),
            ('alpha proportional', stats.alpha_proportional),
            ('phi', stats.phi),
            ('phi average proportional', stats.phi_average_proportional),
        ]
        for name, func in metrics:
            print(name)
            metric = func(pam)
            assert metric.shape == (20,)

    # ............................
    def test_site_tree_stats(self):
        """Test site metrics that take a tree as input."""
        pam, tree = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        metrics = [('Phylogenetic Diversity', stats.phylogenetic_diversity)]
        labels = pam.get_column_headers()
        for name, func in metrics:
            print(name)
            for row in pam:
                present_labels = [labels[i] for i in np.where(row == 1)[0]]
                if present_labels:
                    slice_tree = tree.extract_tree_with_taxa_labels(present_labels)
                    metric = func(slice_tree)
                    assert isinstance(metric, (int, float))

    # ............................
    def test_site_tree_distance_matrix_stats(self):
        """Test site metrics that take a distance matrix as input."""
        pam, tree = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        metrics = [
            ('MNTD', stats.mean_nearest_taxon_distance),
            ('Mean Pairwise Distance', stats.mean_pairwise_distance),
            ('Sum Pairwise Distance', stats.sum_pairwise_distance),
        ]
        labels = pam.get_column_headers()
        for name, func in metrics:
            print(name)
            for row in pam:
                present_labels = [labels[i] for i in np.where(row == 1)[0]]
                if present_labels:
                    slice_tree = tree.extract_tree_with_taxa_labels(present_labels)
                    slice_dist_mtx = slice_tree.get_distance_matrix()
                    metric = func(slice_dist_mtx)
                    assert isinstance(metric, (int, float))

    # ............................
    def test_site_pam_dist_matrix_stats(self):
        """Test site metrics that take a PAM and distance matrix as input."""
        pam, tree = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        dist_mtx = tree.get_distance_matrix()
        metrics = [('pearson_correlation', stats.pearson_correlation)]
        for name, func in metrics:
            print(name)
            metric = func(pam, dist_mtx)
            assert metric.shape == (20, 1)

    # ............................
    def test_species_matrix_metrics(self):
        """Test species metrics."""
        pam, tree = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        metrics = [
            ('omega', stats.omega),
            ('omega_proportional', stats.omega_proportional),
            ('psi', stats.psi),
            ('psi_average_proportional', stats.psi_average_proportional),
        ]
        for name, func in metrics:
            print(name)
            metric = func(pam)
            assert metric.shape == (10,)


# .............................................................................
class Test_PamStats:
    """Test the PamStats class."""

    # ............................
    def test_simple(self):
        """Perform simple PAM stats computations."""
        pam, tree = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        ps.calculate_covariance_statistics()
        ps.calculate_diversity_statistics()
        ps.calculate_site_statistics()
        ps.calculate_species_statistics()

    # ............................
    def test_add_metric(self):
        """Test adding metrics to the statistics computations."""
        pam, tree = get_random_pam_and_tree(10, 20, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        # Remove all metrics
        ps.covariance_stats = []
        ps.diversity_stats = []
        ps.site_matrix_stats = []
        ps.site_tree_stats = []
        ps.site_tree_distance_matrix_stats = []
        ps.site_pam_dist_mtx_stats = []
        ps.species_matrix_stats = []

        # Register all of the metrics
        reg_metrics = [
            ('sigma sites', stats.sigma_sites),
            ('sigma species', stats.sigma_species),
            ('c-score', stats.c_score),
            ('lande', stats.lande),
            ('legendre', stats.legendre),
            ('num sites', stats.num_sites),
            ('num species', stats.num_species),
            ('whittaker', stats.whittaker),
            ('alpha', stats.alpha),
            ('alpha proportional', stats.alpha_proportional),
            ('phi', stats.phi),
            ('phi average proportional', stats.phi_average_proportional),
            ('Phylogenetic Diversity', stats.phylogenetic_diversity),
            ('MNTD', stats.mean_nearest_taxon_distance),
            ('Mean Pairwise Distance', stats.mean_pairwise_distance),
            ('Sum Pairwise Distance', stats.sum_pairwise_distance),
            ('pearson_correlation', stats.pearson_correlation),
            ('omega', stats.omega),
            ('omega_proportional', stats.omega_proportional),
            ('psi', stats.psi),
            ('psi_average_proportional', stats.psi_average_proportional),
        ]
        ps.register_metric('schluter_site_variance', stats.schluter_site_variance_ratio)
        ps.register_metric(
            'schluter_species_variance', stats.schluter_species_variance_ratio
        )
        for name, func in reg_metrics:
            ps.register_metric(name, func)
        ps.calculate_covariance_statistics()
        ps.calculate_diversity_statistics()
        ps.calculate_site_statistics()
        ps.calculate_species_statistics()

        with pytest.raises(TypeError):
            ps.register_metric('bad_metric', int)

    # ............................
    def test_medium_matrix(self):
        """Test species metrics."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        ps.calculate_covariance_statistics()
        ps.calculate_diversity_statistics()
        ps.calculate_site_statistics()
        ps.calculate_species_statistics()

    # ............................
    def test_medium_matrix_with_mismatches(self):
        """Test species metrics."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0, num_mismatches=10)
        ps = stats.PamStats(pam, tree=tree)
        ps.calculate_covariance_statistics()
        ps.calculate_diversity_statistics()
        ps.calculate_site_statistics()
        ps.calculate_species_statistics()

    # ............................
    def test_medium_matrix_with_mismatches_and_empty_row_cols(self):
        """Test species metrics."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0, num_mismatches=10)
        for i in np.random.randint(0, 100, (10,)):
            pam[:, i] = np.zeros((200,))
        for i in np.random.randint(0, 200, (10,)):
            pam[i, :] = np.zeros((100,))
        ps = stats.PamStats(pam, tree=tree)
        ps.calculate_covariance_statistics()
        ps.calculate_diversity_statistics()
        ps.calculate_site_statistics()
        ps.calculate_species_statistics()

    # ............................
    def test_empty_stat_types_site_matrix_stats(self):
        """Test removing all stats for type."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        ps.site_matrix_stats = []
        ps.calculate_site_statistics()

    # ............................
    def test_empty_stat_types_site_tree_stats(self):
        """Test removing all stats for type."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        ps.site_tree_stats = []
        ps.calculate_site_statistics()

    # ............................
    def test_empty_stat_types_site_tree_matrix_stats(self):
        """Test removing all stats for type."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        ps.site_tree_distance_matrix_stats = []
        ps.calculate_site_statistics()

    # ............................
    def test_empty_stat_types_site_pam_dist_matrix_stats(self):
        """Test removing all stats for type."""
        pam, tree = get_random_pam_and_tree(100, 200, 0.3, 1.0)
        ps = stats.PamStats(pam, tree=tree)
        ps.site_pam_dist_mtx_stats = []
        ps.calculate_site_statistics()

    # ............................
    def test_single_double_tree_subsetting(self):
        """Tests for sites with one or two sites present."""
        pam, tree = get_random_pam_and_tree(5, 5, 0.2, 2.0)
        pam[0] = [1, 1, 0, 0, 0]
        pam[1] = [0, 0, 1, 0, 0]
        ps = stats.PamStats(pam, tree=tree)
        ps.calculate_site_statistics()



"""Base init module."""
# flake8: noqa
from . import matrix
from . import randomize
from . import tree
from .matrix import *
from .tree import *
from .point import Point
from . import data_preparation
from . import data_wrangling
from . import spatial
from . import statistics
from . import tools

__all__ = [
    'data_preparation',
    'data_wrangling',
    'randomize',
    'spatial',
    'statistics',
    'tools',
]
__all__.extend(matrix.__all__)
__all__.extend(tree.__all__)

from . import _version

__version__ = _version.get_versions()['version']



"""A tool for wrangling phylogenetic trees."""
import argparse
from copy import deepcopy
import json

from lmpy.data_wrangling.factory import WranglerFactory
from lmpy.tree import TreeWrapper


# .....................................................................................
DESCRIPTION = '''A tool to wrangle phylogenetic trees as needed for processing.'''


# .....................................................................................
def wrangle_tree(tree, wranglers):
    """Wrangle the tree as configured.

    Args:
        tree (TreeWrapper): A phylogenetic tree to wrangle.
        wranglers (list): A list of tree data wranglers.

    Returns:
        TreeWrapper: The modified tree object.
        dict: Wrangler reporting information
    """
    report = []
    for wrangler in wranglers:
        tree = wrangler.wrangle_tree(tree)
        report.append(deepcopy(wrangler.get_report()))
    return tree, report


# .....................................................................................
def cli():
    """Command-line interface for the tool."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '-r', '--report_filename', type=str, help='Path to write report.'
    )
    parser.add_argument('tree_filename', type=str, help='Path to phylogenetic tree.')
    parser.add_argument(
        'tree_schema',
        type=str,
        choices=['newick', 'nexus'],
        help='The schema of the input phylogenetic tree.',
    )
    parser.add_argument(
        'wrangler_configuration_file',
        type=str,
        help='Path to phylogenetic tree wrangler configuration.',
    )
    parser.add_argument(
        'out_tree_filename',
        type=str,
        help='Path to write the wrangled phylogenetic tree.',
    )
    parser.add_argument(
        'out_tree_schema',
        type=str,
        choices=['newick', 'nexus'],
        help='The schema of the output phylogenetic tree.',
    )
    args = parser.parse_args()
    tree = TreeWrapper.get(path=args.tree_filename, schema=args.tree_schema)
    with open(args.wrangler_configuration_file, mode='rt') as in_json:
        wrangler_factory = WranglerFactory()
        wranglers = wrangler_factory.get_wranglers(json.load(in_json))
    wrangled_tree, report = wrangle_tree(tree, wranglers)
    wrangled_tree.write(path=args.out_tree_filename, schema=args.out_tree_schema)
    if args.report_filename:
        with open(args.report_filename, mode='wt') as report_out:
            json.dump(report, report_out)


# .....................................................................................
__all__ = ['cli', 'wrangle_tree']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Convert a lmpy Matrix to a GeoJSON (.geojson) file."""
import argparse
import json

from lmpy.matrix import Matrix
from lmpy.spatial.geojsonify import (
    geojsonify_matrix, geojsonify_matrix_with_shapefile,
)


DESCRIPTION = 'Convert a lmpy Matrix to a GeoJSON file.'


# .....................................................................................
def cli():
    """Provide a command-line tool for converting LMM to GeoJSON."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '--shapefile_filename',
        '-s',
        type=str,
        help=(
            'Path to a shapefile that can be used to generate polygons matching '
            'matrix sites.'
        ),
    )
    parser.add_argument(
        '--resolution',
        '-r',
        type=float,
        help=(
            'Resolution of the polygons in the GeoJSON if a shapefile was not provided.'
            'Otherwise, use points.'
        ),
    )
    parser.add_argument(
        '--omit_value',
        '-o',
        action='append',
        nargs='*',
        type=float,
        help='Properties should be omitted if they have this value for a site.'
    )
    parser.add_argument(
        'in_lmm_filename', type=str, help='Lmpy .lmm filename to convert to GeoJSON.'
    )
    parser.add_argument(
        'out_geojson_filename',
        type=str,
        help='Location to write the converted matrix GeoJSON.',
    )
    args = parser.parse_args()
    mtx = Matrix.load(args.in_lmm_filename)
    if args.shapefile_filename is not None:
        matrix_geojson = geojsonify_matrix_with_shapefile(
            mtx, args.shapefile_filename, omit_values=args.omit_value
        )
    else:
        matrix_geojson = geojsonify_matrix(
            mtx, resolution=args.resolution, omit_values=args.omit_value
        )
    with open(args.out_geojson_filename, mode='wt') as out_json:
        json.dump(matrix_geojson, out_json)


# .....................................................................................
__all__ = ['cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Create a rare species model by intersecting convex hull and ecoregions."""
import argparse
import os
import tempfile

import numpy as np
from osgeo import gdal, gdalconst, ogr, osr

from lmpy.point import PointCsvReader


DESCRIPTION = '''\
Create a "rare species" model by intersecting the convex hull for a group of
occurrences with the ecoregions those points fall within and then rasterizing the
result.'''

# Raster formats
AUTO_FORMAT = 'auto'
ASC_FORMAT = 'AAIGrid'
TIF_FORMAT = 'GTiff'


# .....................................................................................
def create_rare_species_model(
    points,
    ecoregions_filename,
    model_raster_filename,
    raster_format=AUTO_FORMAT,
    nodata_value=-9999,
    burn_value=50,
):
    """Create a rare species model from a convex hull intersected with ecoregions.

    Args:
        points (list of Point): A list of occurrence points to use for model.
        ecoregions_filename (str): The file location for the ecoregions data.
        model_raster_filename (str): The file location to write the model raster.
        raster_format (str): The output raster format (default: auto - use filename).
        nodata_value (int): The value to use for nodata in the model raster.
        burn_value (int): The burn value to use for model presence.
    """
    # Get the desired output raster format, either provided or determine
    raster_format = get_raster_format(raster_format, model_raster_filename)

    # Get ecoregions array
    (
        ecoregion_data,
        min_x,
        min_y,
        max_x,
        max_y,
        cell_size,
        epsg,
    ) = get_ecoregions_array(ecoregions_filename)
    num_rows, num_cols = ecoregion_data.shape

    # Get convex hull array
    val_set = set()
    geom_collection = ogr.Geometry(ogr.wkbGeometryCollection)
    # Process points
    for pt in points:
        # Add geometry to collection
        pt_geom = ogr.Geometry(ogr.wkbPoint)
        pt_geom.AddPoint(pt[0], pt[1])
        geom_collection.AddGeometry(pt_geom)
        # Add to value set
        col = int((pt[0] - min_x) / cell_size)
        row = int((max_y - pt[1]) / cell_size)
        val_set.add(ecoregion_data[row, col])
    # Convex hull
    convex_hull_raw = geom_collection.ConvexHull()

    convex_hull_data = get_convex_hull_array(
        convex_hull_raw, num_cols, num_rows, cell_size, min_x, max_y, epsg, burn_value
    )

    # Create model
    model_data = nodata_value * np.ones(ecoregion_data.shape, dtype=int)
    for i in range(ecoregion_data.shape[0]):
        for j in range(ecoregion_data.shape[1]):
            if ecoregion_data[i, j] in val_set and convex_hull_data[i, j] == burn_value:
                model_data[i, j] = burn_value

    # Write model
    if raster_format == ASC_FORMAT:
        write_ascii(
            model_raster_filename, model_data, cell_size, min_x, min_y, nodata_value
        )
    else:
        write_tiff(
            model_raster_filename,
            model_data,
            cell_size,
            min_x,
            max_y,
            epsg,
            nodata_value,
        )


# .....................................................................................
def get_convex_hull_array(
    convex_hull_geom,
    num_cols,
    num_rows,
    cell_size,
    min_x,
    max_y,
    epsg,
    burn_value,
    buffer_distance=0.5,
    num_quad_segs=30,
):
    """Create a convex hull array.

    Args:
        convex_hull_geom (geometry): Convex hull geometry for a set of points.
        num_cols (int): The number of columns in the array.
        num_rows (int): The number of rows in the array.
        cell_size (number): The size of each cell in map units.
        min_x (number): The minimum x value in the raster.
        max_y (number): The maximum y value in the raster.
        epsg (int): The EPSG code of the map projection to use.
        burn_value (number): A value to burn into the raster for the convex hull.
        buffer_distance (number): A buffer distance to add to the convex hull.
        num_quad_segs (int): The number of segments to use for buffering.

    Returns:
        numpy.ndarray: Convex hull data converted to an array.
    """
    buffered_convex_hull = convex_hull_geom.Buffer(buffer_distance, num_quad_segs)
    tmp_shp_filename = tempfile.NamedTemporaryFile(suffix='.shp', delete=True).name
    tmp_tif_filename = tempfile.NamedTemporaryFile(suffix='.tif', delete=True).name

    shp_drv = ogr.GetDriverByName('ESRI Shapefile')

    # Create the convex hull shapefile
    out_ds = shp_drv.CreateDataSource(tmp_shp_filename)
    out_lyr = out_ds.CreateLayer('ConvexHull', geom_type=ogr.wkbPolygon)

    # Add an ID field
    id_field = ogr.FieldDefn('id', ogr.OFTInteger)
    out_lyr.CreateField(id_field)

    # Create the feature
    feat_defn = out_lyr.GetLayerDefn()
    feat = ogr.Feature(feat_defn)
    feat.SetGeometry(buffered_convex_hull)
    feat.SetField('id', 1)
    out_lyr.CreateFeature(feat)
    feat = None

    # Rasterize the shapefile
    tiff_drv = gdal.GetDriverByName('GTiff')
    rst_ds = tiff_drv.Create(
        tmp_tif_filename, num_cols, num_rows, 1, gdalconst.GDT_Int16
    )
    rst_ds.SetGeoTransform([min_x, cell_size, 0, max_y, 0, -cell_size])

    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)
    rst_ds.SetProjection(srs.ExportToWkt())

    gdal.RasterizeLayer(rst_ds, [1], out_lyr, burn_values=[burn_value])

    rst_ds.FlushCache()

    out_ds = None
    rst_ds = None

    # Get the data array
    rst_ds = gdal.Open(tmp_tif_filename)
    band = rst_ds.GetRasterBand(1)
    data = np.array(band.ReadAsArray())

    return data


# .....................................................................................
def get_ecoregions_array(ecoregions_filename):
    """Get ecoregions data array.

    Args:
        ecoregions_filename (str): File location of ecoregions data.

    Returns:
        tuple: (Ecoregion data as numpy array, min x, min y, max x, max y, cell size)
    """
    region_ds = gdal.Open(ecoregions_filename)
    reg_band = region_ds.GetRasterBand(1)
    num_cols = region_ds.RasterXSize
    num_rows = region_ds.RasterYSize
    min_x, cell_size, _, max_y, _, _ = region_ds.GetGeoTransform()
    min_y = max_y - (cell_size * num_rows)
    max_x = min_x + (num_cols * cell_size)

    proj = osr.SpatialReference(wkt=region_ds.GetProjection())
    epsg = int(proj.GetAttrValue('AUTHORITY', 1))

    ecoregion_data = reg_band.ReadAsArray(0, 0, num_cols, num_rows)
    region_ds = None

    return ecoregion_data, min_x, min_y, max_x, max_y, cell_size, epsg


# .....................................................................................
def get_raster_format(provided_format, model_raster_filename):
    """Get the output raster format.

    Args:
        provided_format (str): The user provided raster format.
        model_raster_filename (str): The model raster output filename.

    Returns:
        str: The output raster format string (ASC_FORMAT or TIF_FORMAT).
    """
    if provided_format in [ASC_FORMAT, TIF_FORMAT]:
        # If provided format is known, use it, else we'll determine it
        return provided_format
    elif os.path.splitext(model_raster_filename)[1].lower() == '.asc':
        # If model filename ends in .asc, use ASCII format
        return ASC_FORMAT
    return TIF_FORMAT


# .....................................................................................
def read_points(csv_filename, sp_key, x_key, y_key):
    """Read points from a csv file.

    Args:
        csv_filename (str): Path to CSV file containing occurrence records.
        sp_key (str): The CSV column header containing species information.
        x_key (str): The CSV column header containing X values.
        y_key (str): The CSV column header containing Y values.

    Returns:
        list of (float, float) tuples: A list of (x, y) tuples for occurrences.
    """
    point_set = set()
    with PointCsvReader(csv_filename, sp_key, x_key, y_key) as reader:
        for points in reader:
            point_set = point_set.union({(pt.x, pt.y) for pt in points})
    return list(point_set)


# .....................................................................................
def write_ascii(out_filename, model_data, cell_size, min_x, min_y, nodata_value):
    """Write model data to ascii file.

    Args:
        out_filename (str): The file location to write the raster.
        model_data (numpy.ndarray): Raster data in the form of a matrix or array.
        cell_size (number): The desired size of each raster cell in map units.
        min_x (number): The minimum x value for the raster in map units.
        min_y (number): The minimum y value for the raster in map units.
        nodata_value (number): The value to be considered "nodata" in the raster.
    """
    with open(out_filename, mode='wt') as out_f:
        out_f.write('ncols      {}\n'.format(model_data.shape[1]))
        out_f.write('nrows      {}\n'.format(model_data.shape[0]))
        out_f.write('xllcorner  {}\n'.format(min_x))
        out_f.write('yllcorner  {}\n'.format(min_y))
        out_f.write('cellsize   {}\n'.format(cell_size))
        out_f.write('NODATA_value       {}\n'.format(nodata_value))
        np.savetxt(out_f, model_data, fmt='%i')


# .....................................................................................
def write_tiff(out_filename, model_data, cell_size, min_x, max_y, epsg, nodata_value):
    """Write model data to tiff file.

    Args:
        out_filename (str): The file location to write the raster.
        model_data (numpy.ndarray): Raster data in the form of a matrix or array.
        cell_size (number): The desired size of each raster cell in map units.
        min_x (number): The minimum x value for the raster in map units.
        max_y (number): The maximum y value for the raster in map units.
        epsg (int): The EPSG code to use to define the map projection.
        nodata_value (number): The value to be considered "nodata" in the raster.
    """
    drv = gdal.GetDriverByName('GTiff')
    dataset = drv.Create(
        out_filename, model_data.shape[1], model_data.shape[0], 1, gdalconst.GDT_Byte
    )
    dataset.SetGeoTransform([min_x, cell_size, 0, max_y, 0, -cell_size])

    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)
    dataset.SetProjection(srs.ExportToWkt())

    out_band = dataset.GetRasterBand(1)
    out_band.WriteArray(model_data)
    out_band.FlushCache()
    out_band.SetNoDataValue(nodata_value)


# .....................................................................................
def cli():
    """Command-line interface for creating a rare species model."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '--species_column',
        '-sp',
        type=str,
        default='species_name',
        help='CSV column for species or group name.',
    )
    parser.add_argument(
        '--x_column',
        '-x',
        type=str,
        default='x',
        help='CSV column containing x coordinate (ex. longitude).',
    )
    parser.add_argument(
        '--y_column',
        '-y',
        type=str,
        default='y',
        help='CSV column containing y coordinate (ex. latitude).',
    )
    parser.add_argument(
        '--burn_value',
        '-b',
        type=int,
        default=50,
        help='Raster burn value for model presence.',
    )
    parser.add_argument(
        '--nodata_value',
        '-n',
        type=int,
        default=-9999,
        help='Raster nodata value for model absence.',
    )
    parser.add_argument(
        '--output_format',
        '-of',
        type=str,
        choices=['auto', 'AAIGrid', 'GTiff'],
        default='auto',
        help=(
            'The output format for the model raster.'
            '(AAIGrid -> Ascii Grid, GTiff -> GeoTiff, auto -> Choose by filename.'
        ),
    )
    parser.add_argument(
        'point_csv_filename', type=str, help='File location of an occurrence csv file.'
    )
    parser.add_argument(
        'ecoregions_filename', type=str, help='File location of ecoregions raster file.'
    )
    parser.add_argument(
        'model_raster_filename',
        type=str,
        help='File location to write the model raster file.',
    )
    args = parser.parse_args()

    # Read points
    points = read_points(
        args.point_csv_filename, args.species_column, args.x_column, args.y_column
    )
    # Create model
    create_rare_species_model(
        points,
        args.ecoregions_filename,
        args.model_raster_filename,
        raster_format=args.output_format,
        nodata_value=args.nodata_value,
        burn_value=args.burn_value,
    )


# .....................................................................................
__all__ = ['cli', 'create_rare_species_model', 'write_tiff']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()
